# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
#
# This module was originally developed as part of the PyUtilib project
# Copyright (c) 2008 Sandia Corporation.
# This software is distributed under the BSD License.
# Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
# the U.S. Government retains certain rights in this software.
# ____________________________________________________________________________________

import gc
import glob
import os
import shutil
import sys
import tempfile
from io import StringIO

from os.path import abspath, dirname

import pyomo.common.unittest as unittest

import pyomo.common.tempfiles as tempfiles

from pyomo.common.dependencies import pyutilib_available
from pyomo.common.log import LoggingIntercept
from pyomo.common.tempfiles import (
    TempfileManager,
    TempfileManagerClass,
    TempfileContextError,
)

old_tempdir = TempfileManager.tempdir
tempdir = None


class Test_LegacyTestSuite(unittest.TestCase):
    def setUp(self):
        global tempdir
        tempdir = tempfile.mkdtemp() + os.sep
        TempfileManager.tempdir = tempdir
        TempfileManager.push()

    def tearDown(self):
        global tempdir
        TempfileManager.pop()
        TempfileManager.tempdir = old_tempdir
        if os.path.exists(tempdir):
            shutil.rmtree(tempdir)
        tempdir = None

    def test_add1(self):
        """Test explicit adding of a file that is missing"""
        try:
            TempfileManager.add_tempfile(tempdir + 'add1')
            self.fail("Expected IOError because file 'add1' does not exist")
        except IOError:
            pass

    def test_add2(self):
        """Test explicit adding of a file that is missing"""
        TempfileManager.add_tempfile(tempdir + 'add2', False)

    def test_add3(self):
        """Test explicit adding of a file that already exists"""
        OUTPUT = open(tempdir + 'add3', 'w')
        OUTPUT.write('tempfile\n')
        OUTPUT.close()
        TempfileManager.add_tempfile(tempdir + 'add3')

    def test_add3_dir(self):
        """Test explicit adding of a directory that already exists"""
        os.mkdir(tempdir + 'add3')
        TempfileManager.add_tempfile(tempdir + 'add3')

    def test_pushpop1(self):
        """Test pushpop logic"""
        TempfileManager.push()
        OUTPUT = open(tempdir + 'pushpop1', 'w')
        OUTPUT.write('tempfile\n')
        OUTPUT.close()
        TempfileManager.add_tempfile(tempdir + 'pushpop1')
        TempfileManager.pop()
        if os.path.exists(tempdir + 'pushpop1'):
            self.fail("pop() failed to clean out files")

    def test_pushpop1_dir(self):
        """Test pushpop logic with directories"""
        TempfileManager.push()
        os.mkdir(tempdir + 'pushpop1')
        TempfileManager.add_tempfile(tempdir + 'pushpop1')
        TempfileManager.pop()
        if os.path.exists(tempdir + 'pushpop1'):
            self.fail("pop() failed to clean out directories")

    def test_pushpop2(self):
        """Test pushpop logic"""
        TempfileManager.push()
        OUTPUT = open(tempdir + 'pushpop2', 'w')
        OUTPUT.write('tempfile\n')
        OUTPUT.close()
        TempfileManager.add_tempfile(tempdir + 'pushpop2')

        TempfileManager.push()
        OUTPUT = open(tempdir + 'pushpop2a', 'w')
        OUTPUT.write('tempfile\n')
        OUTPUT.close()
        TempfileManager.add_tempfile(tempdir + 'pushpop2a')
        TempfileManager.pop()
        if not os.path.exists(tempdir + 'pushpop2'):
            self.fail("pop() clean out all files")
        if os.path.exists(tempdir + 'pushpop2a'):
            self.fail("pop() failed to clean out files")

        TempfileManager.pop()
        if os.path.exists(tempdir + 'pushpop2'):
            self.fail("pop() failed to clean out files")

    def test_pushpop2_dir(self):
        """Test pushpop logic with directories"""
        TempfileManager.push()
        os.mkdir(tempdir + 'pushpop2')
        TempfileManager.add_tempfile(tempdir + 'pushpop2')

        TempfileManager.push()
        os.mkdir(tempdir + 'pushpop2a')
        TempfileManager.add_tempfile(tempdir + 'pushpop2a')
        TempfileManager.pop()
        if not os.path.exists(tempdir + 'pushpop2'):
            self.fail("pop() clean out all files")
        if os.path.exists(tempdir + 'pushpop2a'):
            self.fail("pop() failed to clean out files")

        TempfileManager.pop()
        if os.path.exists(tempdir + 'pushpop2'):
            self.fail("pop() failed to clean out files")

    def test_clear(self):
        """Test clear logic"""
        TempfileManager.push()
        OUTPUT = open(tempdir + 'pushpop2', 'w')
        OUTPUT.write('tempfile\n')
        OUTPUT.close()
        os.mkdir(tempdir + 'pushpopdir2')
        TempfileManager.add_tempfile(tempdir + 'pushpop2')
        TempfileManager.add_tempfile(tempdir + 'pushpopdir2')

        TempfileManager.push()
        OUTPUT = open(tempdir + 'pushpop2a', 'w')
        OUTPUT.write('tempfile\n')
        OUTPUT.close()
        os.mkdir(tempdir + 'pushpopdir2a')
        TempfileManager.add_tempfile(tempdir + 'pushpop2a')
        TempfileManager.add_tempfile(tempdir + 'pushpopdir2a')

        TempfileManager.clear_tempfiles()
        # Push a new context onto the stack so that the tearDown succeeds
        TempfileManager.push()

        if os.path.exists(tempdir + 'pushpop2a'):
            self.fail("clear_tempfiles() failed to clean out files")
        if os.path.exists(tempdir + 'pushpopdir2a'):
            self.fail("clear_tempfiles() failed to clean out directories")
        if os.path.exists(tempdir + 'pushpop2'):
            self.fail("clear_tempfiles() failed to clean out files")
        if os.path.exists(tempdir + 'pushpopdir2'):
            self.fail("clear_tempfiles() failed to clean out directories")

    def test_create1(self):
        """Test create logic - no options"""
        fname = TempfileManager.create_tempfile()
        OUTPUT = open(fname, 'w')
        OUTPUT.write('tempfile\n')
        OUTPUT.close()
        self.assertEqual(len(list(glob.glob(tempdir + '*'))), 1)
        fname = os.path.basename(fname)
        self.assertTrue(fname.startswith('tmp'))

    def test_create1_dir(self):
        """Test create logic - no options"""
        fname = TempfileManager.create_tempdir()
        self.assertEqual(len(list(glob.glob(tempdir + '*'))), 1)
        fname = os.path.basename(fname)
        self.assertTrue(fname.startswith('tmp'))

    def test_create1a(self):
        """Test create logic - no options"""
        fname = TempfileManager.create_tempfile(dir=tempdir)
        OUTPUT = open(fname, 'w')
        OUTPUT.write('tempfile\n')
        OUTPUT.close()
        self.assertEqual(len(list(glob.glob(tempdir + '*'))), 1)
        fname = os.path.basename(fname)
        self.assertTrue(fname.startswith('tmp'))

    def test_create1a_dir(self):
        """Test create logic - no options"""
        fname = TempfileManager.create_tempdir(dir=tempdir)
        self.assertEqual(len(list(glob.glob(tempdir + '*'))), 1)
        fname = os.path.basename(fname)
        self.assertTrue(fname.startswith('tmp'))

    def test_create2(self):
        """Test create logic - no options"""
        fname = TempfileManager.create_tempfile(prefix='foo')
        OUTPUT = open(fname, 'w')
        OUTPUT.write('tempfile\n')
        OUTPUT.close()
        self.assertEqual(len(list(glob.glob(tempdir + '*'))), 1)
        fname = os.path.basename(fname)
        self.assertTrue(fname.startswith('foo'))

    def test_create2_dir(self):
        """Test create logic - no options"""
        fname = TempfileManager.create_tempdir(prefix='foo')
        self.assertEqual(len(list(glob.glob(tempdir + '*'))), 1)
        fname = os.path.basename(fname)
        self.assertTrue(fname.startswith('foo'))

    def test_create3(self):
        """Test create logic - no options"""
        fname = TempfileManager.create_tempfile(suffix='bar')
        OUTPUT = open(fname, 'w')
        OUTPUT.write('tempfile\n')
        OUTPUT.close()
        self.assertEqual(len(list(glob.glob(tempdir + '*'))), 1)
        fname = os.path.basename(fname)
        self.assertTrue(fname.endswith('bar'))

    def test_create3_dir(self):
        """Test create logic - no options"""
        fname = TempfileManager.create_tempdir(suffix='bar')
        self.assertEqual(len(list(glob.glob(tempdir + '*'))), 1)
        fname = os.path.basename(fname)
        self.assertTrue(fname.endswith('bar'))


class Test_TempfileManager(unittest.TestCase):
    def setUp(self):
        self.TM = TempfileManagerClass()

    def tearDown(self):
        self.TM.shutdown()

    def test_create_tempfile(self):
        context = self.TM.push()
        fname = self.TM.create_tempfile("suffix", "prefix")
        self.assertRegex(os.path.basename(fname), '^prefix')
        self.assertRegex(os.path.basename(fname), 'suffix$')
        self.assertGreater(len(os.path.basename(fname)), len('prefixsuffix'))
        self.assertTrue(os.path.exists(fname))
        self.assertTrue(os.path.isfile(fname))
        context.release()
        self.assertFalse(os.path.exists(fname))

    def test_mkstemp(self):
        context = self.TM.new_context()
        fd, fname = context.mkstemp("suffix", "prefix")
        self.assertRegex(os.path.basename(fname), '^prefix')
        self.assertRegex(os.path.basename(fname), 'suffix$')
        self.assertGreater(len(os.path.basename(fname)), len('prefixsuffix'))
        self.assertTrue(os.path.exists(fname))
        self.assertTrue(os.path.isfile(fname))
        # calling fsync should run without error
        os.fsync(fd)
        context.release()
        self.assertFalse(os.path.exists(fname))
        # calling fsync should error, as the file descriptor was closed
        with self.assertRaises(OSError):
            os.fsync(fd)

        context = self.TM.new_context()
        fd, fname = context.mkstemp("suffix", "prefix")
        # closing the fd should not result in an error when the context
        # is released
        os.close(fd)
        context.release()

    def test_mktemp_delete(self):
        with self.TM.new_context() as context:
            dname = context.mkdtemp()
            with self.TM.new_context() as subcontext:
                fd, fname1 = subcontext.mkstemp(dir=dname)
                fd, fname2 = subcontext.mkstemp(dir=dname, delete=False)
                # Note: because the context manager isn't going to
                # delete fname2, we need to explicitly close the file
                # here...
                os.close(fd)
                dname1 = subcontext.mkdtemp(dir=dname)
                dname2 = subcontext.mkdtemp(dir=dname, delete=False)

                self.assertTrue(os.path.exists(fname1))
                self.assertTrue(os.path.exists(fname2))
                self.assertTrue(os.path.exists(dname1))
                self.assertTrue(os.path.exists(dname2))
            self.assertFalse(os.path.exists(fname1))
            self.assertTrue(os.path.exists(fname2))
            self.assertFalse(os.path.exists(dname1))
            self.assertTrue(os.path.exists(dname2))
        self.assertFalse(os.path.exists(fname1))
        self.assertFalse(os.path.exists(fname2))
        self.assertFalse(os.path.exists(dname1))
        self.assertFalse(os.path.exists(dname2))

    def test_create_tempdir(self):
        context = self.TM.push()
        fname = self.TM.create_tempdir("suffix", "prefix")
        self.assertRegex(os.path.basename(fname), '^prefix')
        self.assertRegex(os.path.basename(fname), 'suffix$')
        self.assertGreater(len(os.path.basename(fname)), len('prefixsuffix'))
        self.assertTrue(os.path.exists(fname))
        self.assertTrue(os.path.isdir(fname))
        context.release()
        self.assertFalse(os.path.exists(fname))

    def test_add_tempfile(self):
        context1 = self.TM.push()
        context2 = self.TM.push()
        fname = context1.create_tempfile()
        dname = context1.create_tempdir()
        sub_fname = os.path.join(dname, "testfile")
        self.TM.add_tempfile(fname)
        with self.assertRaisesRegex(
            IOError,
            "Temporary file does not exist: %s" % sub_fname.replace('\\', '\\\\'),
        ):
            self.TM.add_tempfile(sub_fname)
        self.TM.add_tempfile(sub_fname, exists=False)
        with open(sub_fname, "w") as FILE:
            FILE.write("\n")

        self.assertTrue(os.path.exists(fname))
        self.assertTrue(os.path.exists(dname))
        self.assertTrue(os.path.exists(sub_fname))
        self.TM.pop()
        self.assertFalse(os.path.exists(fname))
        self.assertTrue(os.path.exists(dname))
        self.assertFalse(os.path.exists(sub_fname))

        # Releasing a context that is missing files should be OK (no error)
        self.TM.pop()
        self.assertFalse(os.path.exists(fname))
        self.assertFalse(os.path.exists(dname))
        self.assertFalse(os.path.exists(sub_fname))

    def test_sequential_files(self):
        with LoggingIntercept() as LOG:
            self.assertIsNone(self.TM.sequential_files())
        self.assertIn(
            "The TempfileManager.sequential_files() method has been removed",
            LOG.getvalue().replace('\n', ' '),
        )
        self.assertIsNone(self.TM.unique_files())

    def test_gettempprefix(self):
        ctx = self.TM.new_context()
        pre = ctx.gettempprefix()
        self.assertIsInstance(pre, str)
        self.assertEqual(pre, tempfile.gettempprefix())

        preb = ctx.gettempprefixb()
        self.assertIsInstance(preb, bytes)
        self.assertEqual(preb, tempfile.gettempprefixb())

    def test_gettempdir(self):
        context = self.TM.push()
        # Creating in the system TMP
        fname = context.create_tempfile()
        self.assertIsInstance(fname, str)
        system_tmpdir = os.path.dirname(fname)
        self.assertEqual(system_tmpdir, tempfile.gettempdir())
        tmpdir = context.gettempdir()
        self.assertIsInstance(tmpdir, str)
        self.assertEqual(tmpdir, system_tmpdir)
        tmpdirb = context.gettempdirb()
        self.assertIsInstance(tmpdirb, bytes)
        self.assertEqual(tmpdirb.decode(), tmpdir)

        # Creating in a TMP specified on the TempfileManager
        manager_tmpdir = context.create_tempdir()
        self.assertNotEqual(manager_tmpdir, system_tmpdir)
        self.TM.tempdir = manager_tmpdir
        fname = context.create_tempfile()
        self.assertIsInstance(fname, str)
        tmpdir = context.gettempdir()
        self.assertIsInstance(tmpdir, str)
        self.assertEqual(tmpdir, manager_tmpdir)
        tmpdirb = context.gettempdirb()
        self.assertIsInstance(tmpdirb, bytes)
        self.assertEqual(tmpdirb.decode(), tmpdir)

        # Creating in a TMP specified on the context
        context_tmpdir = context.create_tempdir()
        self.assertNotEqual(context_tmpdir, system_tmpdir)
        self.assertNotEqual(context_tmpdir, manager_tmpdir)
        context.tempdir = context_tmpdir
        fname = context.create_tempfile()
        self.assertIsInstance(fname, str)
        tmpdir = context.gettempdir()
        self.assertIsInstance(tmpdir, str)
        self.assertEqual(tmpdir, context_tmpdir)
        tmpdirb = context.gettempdirb()
        self.assertIsInstance(tmpdirb, bytes)
        self.assertEqual(tmpdirb.decode(), tmpdir)

        # Creating in a TMP specified on the context ... but in bytes
        context.tempdir = context_tmpdir.encode()
        fname = context.create_tempfile()
        self.assertIsInstance(fname, bytes)
        tmpdir = context.gettempdir()
        self.assertIsInstance(tmpdir, str)
        self.assertEqual(tmpdir, context_tmpdir)
        tmpdirb = context.gettempdirb()
        self.assertIsInstance(tmpdirb, bytes)
        self.assertEqual(tmpdirb.decode(), tmpdir)
        # Cleanup
        self.TM.pop()

    def test_shutdown(self):
        with LoggingIntercept() as LOG:
            self.TM.shutdown()
        self.assertEqual(LOG.getvalue(), "")

        self.TM = TempfileManagerClass()
        ctx = self.TM.push()
        with LoggingIntercept() as LOG:
            self.TM.shutdown()
        self.assertEqual(
            LOG.getvalue().strip(),
            "TempfileManagerClass instance: un-popped tempfile "
            "contexts still exist during TempfileManager instance "
            "shutdown",
        )

        self.TM = TempfileManagerClass()
        ctx = self.TM.push()
        fname = ctx.create_tempfile()
        self.assertTrue(os.path.exists(fname))
        with LoggingIntercept() as LOG:
            self.TM.shutdown()
        self.assertFalse(os.path.exists(fname))
        self.assertEqual(
            LOG.getvalue().strip(),
            "Temporary files created through TempfileManager "
            "contexts have not been deleted (observed during "
            "TempfileManager instance shutdown).\n"
            "Undeleted entries:\n\t%s\n"
            "TempfileManagerClass instance: un-popped tempfile "
            "contexts still exist during TempfileManager instance "
            "shutdown" % fname,
        )

        # The TM is already shut down, so this should be a noop
        with LoggingIntercept() as LOG:
            self.TM.shutdown()
        self.assertEqual(LOG.getvalue(), "")

    def test_del_clears_contexts(self):
        TM = TempfileManagerClass()
        ctx = TM.push()
        fname = ctx.create_tempfile()
        self.assertTrue(os.path.exists(fname))
        with LoggingIntercept() as LOG:
            TM = None
            gc.collect()
            gc.collect()
            gc.collect()
        self.assertFalse(os.path.exists(fname))
        self.assertEqual(
            LOG.getvalue().strip(),
            "Temporary files created through TempfileManager "
            "contexts have not been deleted (observed during "
            "TempfileManager instance shutdown).\n"
            "Undeleted entries:\n\t%s\n"
            "TempfileManagerClass instance: un-popped tempfile "
            "contexts still exist during TempfileManager instance "
            "shutdown" % fname,
        )

    def test_tempfilemanager_as_context_manager(self):
        with LoggingIntercept() as LOG:
            with self.TM:
                fname = self.TM.create_tempfile()
                self.assertTrue(os.path.exists(fname))
            self.assertFalse(os.path.exists(fname))
            self.assertEqual(LOG.getvalue(), "")

            with self.TM:
                self.TM.push()
                fname = self.TM.create_tempfile()
                self.assertTrue(os.path.exists(fname))
            self.assertFalse(os.path.exists(fname))
            self.assertEqual(
                LOG.getvalue().strip(),
                "TempfileManager: tempfile context was pushed onto "
                "the TempfileManager stack within a context manager "
                "(i.e., `with TempfileManager:`) but was not popped "
                "before the context manager exited.  Popping the "
                "context to preserve the stack integrity.",
            )

    def test_tempfilecontext_as_context_manager(self):
        with LoggingIntercept() as LOG:
            ctx = self.TM.new_context()
            with ctx:
                fname = ctx.create_tempfile()
                self.assertTrue(os.path.exists(fname))
            self.assertFalse(os.path.exists(fname))
            self.assertEqual(LOG.getvalue(), "")

    @unittest.skipIf(
        not sys.platform.lower().startswith('win'),
        "test only applies to Windows platforms",
    )
    def test_open_tempfile_windows(self):
        self.TM.push()
        fname = self.TM.create_tempfile()
        f = open(fname)
        try:
            _orig = tempfiles.deletion_errors_are_fatal
            tempfiles.deletion_errors_are_fatal = True
            with self.assertRaisesRegex(
                WindowsError, ".*process cannot access the file"
            ):
                self.TM.pop()
        finally:
            tempfiles.deletion_errors_are_fatal = _orig
            f.close()
            os.remove(fname)

        self.TM.push()
        fname = self.TM.create_tempfile()
        f = open(fname)
        try:
            _orig = tempfiles.deletion_errors_are_fatal
            tempfiles.deletion_errors_are_fatal = False
            with LoggingIntercept(None, 'pyomo.common') as LOG:
                self.TM.pop()
            self.assertIn("Unable to delete temporary file", LOG.getvalue())
        finally:
            tempfiles.deletion_errors_are_fatal = _orig
            f.close()
            os.remove(fname)

    @unittest.skipUnless(pyutilib_available, "deprecation test requires pyutilib")
    def test_deprecated_tempdir(self):
        self.TM.push()
        try:
            tmpdir = self.TM.create_tempdir()
            _orig = tempfiles.pyutilib_tempfiles.TempfileManager.tempdir
            tempfiles.pyutilib_tempfiles.TempfileManager.tempdir = tmpdir
            self.TM.tempdir = None

            with LoggingIntercept() as LOG:
                fname = self.TM.create_tempfile()
            self.assertIn(
                "The use of the PyUtilib TempfileManager.tempdir "
                "to specify the default location for Pyomo "
                "temporary files",
                LOG.getvalue().replace("\n", " "),
            )

            with LoggingIntercept() as LOG:
                dname = self.TM.create_tempdir()
            self.assertIn(
                "The use of the PyUtilib TempfileManager.tempdir "
                "to specify the default location for Pyomo "
                "temporary files",
                LOG.getvalue().replace("\n", " "),
            )
        finally:
            self.TM.pop()
            tempfiles.pyutilib_tempfiles.TempfileManager.tempdir = _orig

    def test_context(self):
        with self.assertRaisesRegex(
            TempfileContextError, "TempfileManager has no currently active context"
        ):
            self.TM.context()
        ctx = self.TM.push()
        self.assertIs(ctx, self.TM.context())


if __name__ == "__main__":
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import copy

from io import StringIO
from pyomo.core.expr import expr_common


import pyomo.common.unittest as unittest
import pyomo.core.expr as EXPR

from pyomo.environ import (
    ConcreteModel,
    AbstractModel,
    Expression,
    Var,
    Set,
    Param,
    Objective,
    value,
    sum_product,
)
from pyomo.core.base.expression import ExpressionData
from pyomo.core.base.objective import ObjectiveData
from pyomo.core.expr.compare import compare_expressions, assertExpressionsEqual
from pyomo.common.tee import capture_output


class TestExpressionData(unittest.TestCase):
    def test_exprdata_get_set(self):
        model = ConcreteModel()
        model.e = Expression([1])
        self.assertEqual(len(model.e), 1)
        self.assertEqual(model.e[1].expr, None)
        model.e.add(1, 1)
        self.assertEqual(model.e[1].expr, 1)
        model.e[1].expr += 2
        self.assertEqual(model.e[1].expr, 3)

    def test_exprdata_get_set_value(self):
        model = ConcreteModel()
        model.e = Expression([1])
        self.assertEqual(len(model.e), 1)
        self.assertEqual(model.e[1].expr, None)
        model.e.add(1, 1)
        model.e[1].expr = 1
        self.assertEqual(model.e[1].expr, 1)
        model.e[1].expr += 2
        self.assertEqual(model.e[1].expr, 3)

    # The copy method must be invoked on expression container to obtain
    # a shallow copy of the class, the underlying expression remains
    # a reference.
    def test_copy(self):
        model = ConcreteModel()
        model.a = Var(initialize=5)
        model.b = Var(initialize=10)

        model.expr1 = Expression(initialize=model.a)

        # Do a shallow copy, the same underlying expression is still referenced
        expr2 = copy.copy(model.expr1)
        self.assertEqual(model.expr1(), 5)
        self.assertEqual(expr2(), 5)
        self.assertEqual(id(model.expr1.expr), id(expr2.expr))

        # Do an in place modification the expression
        model.expr1.expr.set_value(1)
        self.assertEqual(model.expr1(), 1)
        self.assertEqual(expr2(), 1)
        self.assertEqual(id(model.expr1.expr), id(expr2.expr))

        # Update the expression value on expr1 only
        model.expr1.set_value(model.b)
        self.assertEqual(model.expr1(), 10)
        self.assertEqual(expr2(), 1)
        self.assertNotEqual(id(model.expr1.expr), id(expr2.expr))

        model.a.set_value(5)
        model.b.set_value(10)
        model.del_component('expr1')
        model.expr1 = Expression(initialize=model.a + model.b)

        # Do a shallow copy, the same underlying expression is still referenced
        expr2 = copy.copy(model.expr1)
        self.assertEqual(model.expr1(), 15)
        self.assertEqual(expr2(), 15)
        self.assertEqual(id(model.expr1.expr), id(expr2.expr))
        self.assertEqual(id(model.expr1.expr.arg(0)), id(expr2.expr.arg(0)))
        self.assertEqual(id(model.expr1.expr.arg(1)), id(expr2.expr.arg(1)))

        # Do an in place modification the expression
        # This causes cloning due to reference counting
        model.a.set_value(0)
        self.assertEqual(model.expr1(), 10)
        self.assertEqual(expr2(), 10)
        self.assertEqual(id(model.expr1.expr), id(expr2.expr))
        self.assertEqual(id(model.expr1.expr.arg(0)), id(expr2.expr.arg(0)))
        self.assertEqual(id(model.expr1.expr.arg(1)), id(expr2.expr.arg(1)))

        # Do an in place modification the expression
        # This causes cloning due to reference counting
        model.expr1.expr += 1
        self.assertEqual(model.expr1(), 11)
        self.assertEqual(expr2(), 10)
        self.assertNotEqual(id(model.expr1.expr), id(expr2.expr))

    # test that an object is properly deepcopied when the model is cloned
    def test_model_clone(self):
        model = ConcreteModel()
        model.x = Var(initialize=2.0)
        model.y = Var(initialize=0.0)
        model.ec = Expression(initialize=model.x**2 + 1)
        model.obj = Objective(expr=model.y + model.ec)
        self.assertEqual(model.obj.expr(), 5.0)
        self.assertTrue(id(model.ec) in [id(e) for e in model.obj.expr.args])
        inst = model.clone()
        self.assertEqual(inst.obj.expr(), 5.0)
        if not id(inst.ec) in [id(e) for e in inst.obj.expr.args]:
            print("BUG?")
            print(id(inst.ec))
            print(inst.obj.expr.__class__)
            print([id(e) for e in inst.obj.expr.args])
            print([e.__class__ for e in inst.obj.expr.args])
            print([id(e) for e in model.obj.expr.args])
            print([e.__class__ for e in model.obj.expr.args])
        self.assertTrue(id(inst.ec) in [id(e) for e in inst.obj.expr.args])
        self.assertNotEqual(id(model.ec), id(inst.ec))
        self.assertFalse(id(inst.ec) in [id(e) for e in model.obj.expr.args])

    def test_is_constant(self):
        model = ConcreteModel()
        model.x = Var(initialize=1.0)
        model.p = Param(initialize=1.0)
        model.ec = Expression(initialize=model.x)
        self.assertEqual(model.ec.is_constant(), False)
        self.assertEqual(model.ec.expr.is_constant(), False)
        model.ec.set_value(model.p)
        self.assertEqual(model.ec.is_constant(), False)
        self.assertEqual(model.ec.expr.is_constant(), True)

    def test_polynomial_degree(self):
        model = ConcreteModel()
        model.x = Var(initialize=1.0)
        model.ec = Expression(initialize=model.x)
        self.assertEqual(
            model.ec.polynomial_degree(), model.ec.expr.polynomial_degree()
        )
        self.assertEqual(model.ec.polynomial_degree(), 1)
        model.ec.set_value(model.x**2)
        self.assertEqual(
            model.ec.polynomial_degree(), model.ec.expr.polynomial_degree()
        )
        self.assertEqual(model.ec.polynomial_degree(), 2)

    def test_init_concrete(self):
        model = ConcreteModel()
        model.y = Var(initialize=0.0)
        model.x = Var(initialize=1.0)

        model.ec = Expression(expr=0)
        model.obj = Objective(expr=1.0 + model.ec)
        self.assertEqual(model.obj.expr(), 1.0)
        self.assertEqual(id(model.obj.expr.arg(1)), id(model.ec))
        e = 1.0
        model.ec.set_value(e)
        self.assertEqual(model.obj.expr(), 2.0)
        self.assertEqual(id(model.obj.expr.arg(1)), id(model.ec))
        e += model.x
        model.ec.set_value(e)
        self.assertEqual(model.obj.expr(), 3.0)
        self.assertEqual(id(model.obj.expr.arg(1)), id(model.ec))
        e += model.x
        self.assertEqual(model.obj.expr(), 3.0)
        self.assertEqual(id(model.obj.expr.arg(1)), id(model.ec))

        model.del_component('obj')
        model.del_component('ec')
        model.ec = Expression(initialize=model.y)
        model.obj = Objective(expr=1.0 + model.ec)
        self.assertEqual(model.obj.expr(), 1.0)
        self.assertEqual(id(model.obj.expr.arg(1)), id(model.ec))
        e = 1.0
        model.ec.set_value(e)
        self.assertEqual(model.obj.expr(), 2.0)
        self.assertEqual(id(model.obj.expr.arg(1)), id(model.ec))
        e += model.x
        model.ec.set_value(e)
        self.assertEqual(model.obj.expr(), 3.0)
        self.assertEqual(id(model.obj.expr.arg(1)), id(model.ec))
        e += model.x
        self.assertEqual(model.obj.expr(), 3.0)
        self.assertEqual(id(model.obj.expr.arg(1)), id(model.ec))

        model.del_component('obj')
        model.del_component('ec')
        model.y.set_value(-1)
        model.ec = Expression(initialize=model.y + 1.0)
        model.obj = Objective(expr=1.0 + model.ec)
        self.assertEqual(model.obj.expr(), 1.0)
        self.assertEqual(id(model.obj.expr.arg(1)), id(model.ec))
        e = 1.0
        model.ec.set_value(e)
        self.assertEqual(model.obj.expr(), 2.0)
        self.assertEqual(id(model.obj.expr.arg(1)), id(model.ec))
        e += model.x
        model.ec.set_value(e)
        self.assertEqual(model.obj.expr(), 3.0)
        self.assertEqual(id(model.obj.expr.arg(1)), id(model.ec))
        e += model.x
        self.assertEqual(model.obj.expr(), 3.0)
        self.assertEqual(id(model.obj.expr.arg(1)), id(model.ec))

    def test_init_abstract(self):
        model = AbstractModel()
        model.y = Var(initialize=0.0)
        model.x = Var(initialize=1.0)
        model.ec = Expression(initialize=0.0)

        def obj_rule(model):
            return 1.0 + model.ec

        model.obj = Objective(rule=obj_rule)
        inst = model.create_instance()
        self.assertEqual(inst.obj.expr(), 1.0)
        self.assertEqual(id(inst.obj.expr.arg(1)), id(inst.ec))
        e = 1.0
        inst.ec.set_value(e)
        self.assertEqual(inst.obj.expr(), 2.0)
        self.assertEqual(id(inst.obj.expr.arg(1)), id(inst.ec))
        e += inst.x
        inst.ec.set_value(e)
        self.assertEqual(inst.obj.expr(), 3.0)
        self.assertEqual(id(inst.obj.expr.arg(1)), id(inst.ec))
        e += inst.x
        self.assertEqual(inst.obj.expr(), 3.0)
        self.assertEqual(id(inst.obj.expr.arg(1)), id(inst.ec))

        model.del_component('obj')
        model.del_component('ec')
        model.ec = Expression(initialize=0.0)

        def obj_rule(model):
            return 1.0 + model.ec

        model.obj = Objective(rule=obj_rule)
        inst = model.create_instance()
        self.assertEqual(inst.obj.expr(), 1.0)
        self.assertEqual(id(inst.obj.expr.arg(1)), id(inst.ec))
        e = 1.0
        inst.ec.set_value(e)
        self.assertEqual(inst.obj.expr(), 2.0)
        self.assertEqual(id(inst.obj.expr.arg(1)), id(inst.ec))
        e += inst.x
        inst.ec.set_value(e)
        self.assertEqual(inst.obj.expr(), 3.0)
        self.assertEqual(id(inst.obj.expr.arg(1)), id(inst.ec))
        e += inst.x
        self.assertEqual(inst.obj.expr(), 3.0)
        self.assertEqual(id(inst.obj.expr.arg(1)), id(inst.ec))

        model.del_component('obj')
        model.del_component('ec')
        model.ec = Expression(initialize=0.0)

        def obj_rule(model):
            return 1.0 + model.ec

        model.obj = Objective(rule=obj_rule)
        inst = model.create_instance()
        self.assertEqual(inst.obj.expr(), 1.0)
        self.assertEqual(id(inst.obj.expr.arg(1)), id(inst.ec))
        e = 1.0
        inst.ec.set_value(e)
        self.assertEqual(inst.obj.expr(), 2.0)
        self.assertEqual(id(inst.obj.expr.arg(1)), id(inst.ec))
        e += inst.x
        inst.ec.set_value(e)
        self.assertEqual(inst.obj.expr(), 3.0)
        self.assertEqual(id(inst.obj.expr.arg(1)), id(inst.ec))
        e += inst.x
        self.assertEqual(inst.obj.expr(), 3.0)
        self.assertEqual(id(inst.obj.expr.arg(1)), id(inst.ec))

    def test_create_node_with_local_data(self):
        m = ConcreteModel()
        m.x = Var()

        m.e = Expression(expr=m.x)
        ee = m.e.create_node_with_local_data([5])
        self.assertIsNot(m.e, ee)
        self.assertIs(type(ee), ExpressionData)
        self.assertEqual(ee._args_, [5])

        m.f = Expression([0], rule=lambda m, i: m.x)
        ff = m.f[0].create_node_with_local_data([5])
        self.assertIsNot(m.f, ff)
        self.assertIsNot(m.f[0], ff)
        self.assertIs(type(ff), ExpressionData)
        self.assertEqual(ff._args_, [5])

        m.g = Objective(expr=m.x)
        gg = m.g.create_node_with_local_data([5])
        self.assertIsNot(m.g, gg)
        self.assertIs(type(gg), ObjectiveData)
        self.assertEqual(gg._args_, [5])

        m.h = Objective([0], rule=lambda m, i: m.x)
        hh = m.h[0].create_node_with_local_data([5])
        self.assertIsNot(m.h, hh)
        self.assertIsNot(m.h[0], hh)
        self.assertIs(type(hh), ObjectiveData)
        self.assertEqual(hh._args_, [5])


class TestExpression(unittest.TestCase):
    def setUp(self):
        TestExpression._save = expr_common.TO_STRING_VERBOSE
        # Tests can choose what they want - this just makes sure that
        # things are restored after the tests run.
        # expr_common.TO_STRING_VERBOSE = True

    def tearDown(self):
        expr_common.TO_STRING_VERBOSE = TestExpression._save

    def test_unconstructed_singleton(self):
        a = Expression()
        self.assertEqual(a._constructed, False)
        self.assertEqual(len(a), 0)
        try:
            a()
            self.fail("Component is unconstructed")
        except ValueError:
            pass
        try:
            a.expr
            self.fail("Component is unconstructed")
        except ValueError:
            pass
        try:
            a.is_constant()
            self.fail("Component is unconstructed")
        except ValueError:
            pass
        try:
            a.is_fixed()
            self.fail("Component is unconstructed")
        except ValueError:
            pass
        try:
            a.set_value(4)
            self.fail("Component is unconstructed")
        except ValueError:
            pass
        a.construct()
        self.assertEqual(len(a), 1)
        self.assertEqual(a(), None)
        self.assertEqual(a.expr, None)
        self.assertEqual(a.is_constant(), False)
        a.set_value(5)
        self.assertEqual(len(a), 1)
        self.assertEqual(a(), 5)
        self.assertEqual(a.expr, 5)
        self.assertEqual(a.is_constant(), False)
        self.assertEqual(a.is_fixed(), True)

    def test_display(self):
        model = ConcreteModel()
        model.e = Expression()
        with capture_output() as out:
            model.e.display()
        self.assertEqual(
            out.getvalue().strip(),
            """
e : Size=1
    Key  : Value
    None : Undefined
        """.strip(),
        )

        model.e.set_value(1.0)
        with capture_output() as out:
            model.e.display()
        self.assertEqual(
            out.getvalue().strip(),
            """
e : Size=1
    Key  : Value
    None :   1.0
        """.strip(),
        )

        out = StringIO()
        with capture_output() as no_out:
            model.e.display(ostream=out)
        self.assertEqual(no_out.getvalue(), "")
        self.assertEqual(
            out.getvalue().strip(),
            """
e : Size=1
    Key  : Value
    None :   1.0
        """.strip(),
        )

        model.E = Expression([1, 2])
        with capture_output() as out:
            model.E.display()
        self.assertEqual(
            out.getvalue().strip(),
            """
E : Size=2
    Key : Value
      1 : Undefined
      2 : Undefined
        """.strip(),
        )

        model.E[1].set_value(1.0)
        with capture_output() as out:
            model.E.display()
        self.assertEqual(
            out.getvalue().strip(),
            """
E : Size=2
    Key : Value
      1 :       1.0
      2 : Undefined
        """.strip(),
        )

        out = StringIO()
        with capture_output() as no_out:
            model.E.display(ostream=out)
        self.assertEqual(no_out.getvalue(), "")
        self.assertEqual(
            out.getvalue().strip(),
            """
E : Size=2
    Key : Value
      1 :       1.0
      2 : Undefined
        """.strip(),
        )

    def test_extract_values_store_values(self):
        model = ConcreteModel()
        model.e = Expression()
        self.assertEqual(model.e.extract_values(), {None: None})
        model.e.store_values({None: 1.0})
        self.assertEqual(model.e.extract_values(), {None: 1.0})
        with self.assertRaises(KeyError):
            model.e.store_values({1: 1.0})

        model.E = Expression([1, 2])
        self.assertEqual(model.E.extract_values(), {1: None, 2: None})
        model.E.store_values({1: 1.0})
        self.assertEqual(model.E.extract_values(), {1: 1.0, 2: None})
        model.E.store_values({1: None, 2: 2.0})
        self.assertEqual(model.E.extract_values(), {1: None, 2: 2.0})
        with self.assertRaises(KeyError):
            model.E.store_values({3: 3.0})

    def test_setitem(self):
        model = ConcreteModel()
        model.E = Expression([1])
        model.E[1] = 1
        self.assertEqual(value(model.E[1]), 1)
        with self.assertRaises(KeyError):
            model.E[2] = 1
        model.del_component(model.E)
        model.Index = Set(dimen=3, initialize=[(1, 2, 3)])
        model.E = Expression(model.Index)
        model.E[(1, 2, 3)] = 1
        self.assertEqual(value(model.E[(1, 2, 3)]), 1)
        # GH: testing this ludicrous behavior simply for
        #     coverage in expression.py.
        model.E[(1, (2, 3))] = 1
        self.assertEqual(value(model.E[(1, 2, 3)]), 1)
        with self.assertRaises(KeyError):
            model.E[2] = 1

    def test_nonindexed_construct_rule(self):
        model = ConcreteModel()

        def _some_rule(model):
            return 1.0

        model.e = Expression(rule=_some_rule)
        self.assertEqual(value(model.e), 1.0)
        model.del_component(model.e)
        del _some_rule

        def _some_rule(model):
            return Expression.Skip

        model.e = Expression(rule=_some_rule)
        self.assertEqual(len(model.e), 0)

    def test_nonindexed_construct_expr(self):
        model = ConcreteModel()
        model.e = Expression(expr=Expression.Skip)
        self.assertEqual(len(model.e), 0)
        model.del_component(model.e)
        model.e = Expression()
        self.assertEqual(model.e.extract_values(), {None: None})
        model.del_component(model.e)
        model.e = Expression(expr=1.0)
        self.assertEqual(model.e.extract_values(), {None: 1.0})
        model.del_component(model.e)
        model.e = Expression(expr={None: 1.0})
        self.assertEqual(model.e.extract_values(), {None: 1.0})
        # Even though add can be called with any
        # indexed on indexed Expressions, None must
        # always be used as the index for non-indexed
        # Expressions
        with self.assertRaises(KeyError):
            model.e.add(2, 2)

    def test_indexed_construct_rule(self):
        model = ConcreteModel()
        model.Index = Set(initialize=[1, 2, 3])

        def _some_rule(model, i):
            if i == 1:
                return Expression.Skip
            else:
                return i

        model.E = Expression(model.Index, rule=_some_rule)
        self.assertEqual(model.E.extract_values(), {2: 2, 3: 3})
        self.assertEqual(len(model.E), 2)

    def test_implicit_definition(self):
        model = ConcreteModel()
        model.idx = Set(initialize=[1, 2, 3])
        model.E = Expression(model.idx)
        self.assertEqual(len(model.E), 3)
        expr = model.E[1]
        self.assertIs(type(expr), ExpressionData)
        model.E[1] = None
        self.assertIs(expr, model.E[1])
        self.assertIs(type(expr), ExpressionData)
        self.assertIs(expr.expr, None)
        model.E[1] = 5
        self.assertIs(expr, model.E[1])
        self.assertEqual(model.E.extract_values(), {1: 5, 2: None, 3: None})
        model.E[2] = 6
        self.assertIsNot(expr, model.E[2])
        self.assertEqual(model.E.extract_values(), {1: 5, 2: 6, 3: None})

    def test_explicit_skip_definition(self):
        model = ConcreteModel()
        model.idx = Set(initialize=[1, 2, 3])
        model.E = Expression(model.idx, rule=lambda m, i: Expression.Skip)
        self.assertEqual(len(model.E), 0)
        with self.assertRaises(KeyError):
            expr = model.E[1]

        model.E[1] = None
        expr = model.E[1]
        self.assertIs(type(expr), ExpressionData)
        self.assertIs(expr.expr, None)
        model.E[1] = 5
        self.assertIs(expr, model.E[1])
        self.assertEqual(model.E.extract_values(), {1: 5})
        model.E[2] = 6
        self.assertIsNot(expr, model.E[2])
        self.assertEqual(model.E.extract_values(), {1: 5, 2: 6})

    def test_indexed_construct_expr(self):
        model = ConcreteModel()
        model.Index = Set(initialize=[1, 2, 3])
        model.E = Expression(model.Index, expr=Expression.Skip)
        self.assertEqual(len(model.E), 0)
        model.E = Expression(model.Index)
        self.assertEqual(model.E.extract_values(), {1: None, 2: None, 3: None})
        model.del_component(model.E)
        model.E = Expression(model.Index, expr=1.0)
        self.assertEqual(model.E.extract_values(), {1: 1.0, 2: 1.0, 3: 1.0})
        model.del_component(model.E)
        model.E = Expression(
            model.Index, expr={1: Expression.Skip, 2: Expression.Skip, 3: 1.0}
        )
        self.assertEqual(model.E.extract_values(), {3: 1.0})

    def test_bad_init_too_many_keywords(self):
        model = ConcreteModel()

        def _some_rule(model):
            return 1.0

        with self.assertRaises(ValueError):
            model.e = Expression(expr=1.0, rule=_some_rule)
        del _some_rule

        def _some_indexed_rule(model, i):
            return 1.0

        with self.assertRaises(ValueError):
            model.e = Expression([1], expr=1.0, rule=_some_indexed_rule)
        del _some_indexed_rule

    def test_init_concrete_indexed(self):
        model = ConcreteModel()
        model.y = Var(initialize=0.0)
        model.x = Var([1, 2, 3], initialize=1.0)

        model.ec = Expression([1, 2, 3], initialize=1.0)
        model.obj = Objective(expr=1.0 + sum_product(model.ec, index=[1, 2, 3]))
        self.assertEqual(model.obj.expr(), 4.0)
        model.ec[1].set_value(2.0)
        self.assertEqual(model.obj.expr(), 5.0)

    def test_init_concrete_nonindexed(self):
        model = ConcreteModel()
        model.y = Var(initialize=0.0)
        model.x = Var(initialize=1.0)

        model.ec = Expression(initialize=0)
        model.obj = Objective(expr=1.0 + model.ec)
        self.assertEqual(model.obj.expr(), 1.0)
        self.assertEqual(id(model.obj.expr.arg(1)), id(model.ec))
        e = 1.0
        model.ec.set_value(e)
        self.assertEqual(model.obj.expr(), 2.0)
        self.assertEqual(id(model.obj.expr.arg(1)), id(model.ec))
        e += model.x
        model.ec.set_value(e)
        self.assertEqual(model.obj.expr(), 3.0)
        self.assertEqual(id(model.obj.expr.arg(1)), id(model.ec))
        e += model.x
        self.assertEqual(model.obj.expr(), 3.0)
        self.assertEqual(id(model.obj.expr.arg(1)), id(model.ec))

        model.del_component('obj')
        model.del_component('ec')
        model.ec = Expression(initialize=model.y)
        model.obj = Objective(expr=1.0 + model.ec)
        self.assertEqual(model.obj.expr(), 1.0)
        self.assertEqual(id(model.obj.expr.arg(1)), id(model.ec))
        e = 1.0
        model.ec.set_value(e)
        self.assertEqual(model.obj.expr(), 2.0)
        self.assertEqual(id(model.obj.expr.arg(1)), id(model.ec))
        e += model.x
        model.ec.set_value(e)
        self.assertEqual(model.obj.expr(), 3.0)
        self.assertEqual(id(model.obj.expr.arg(1)), id(model.ec))
        e += model.x
        self.assertEqual(model.obj.expr(), 3.0)
        self.assertEqual(id(model.obj.expr.arg(1)), id(model.ec))

        model.del_component('obj')
        model.del_component('ec')
        model.y.set_value(-1)
        model.ec = Expression(initialize=model.y + 1.0)
        model.obj = Objective(expr=1.0 + model.ec)
        self.assertEqual(model.obj.expr(), 1.0)
        self.assertEqual(id(model.obj.expr.arg(1)), id(model.ec))
        e = 1.0
        model.ec.set_value(e)
        self.assertEqual(model.obj.expr(), 2.0)
        self.assertEqual(id(model.obj.expr.arg(1)), id(model.ec))
        e += model.x
        model.ec.set_value(e)
        self.assertEqual(model.obj.expr(), 3.0)
        self.assertEqual(id(model.obj.expr.arg(1)), id(model.ec))
        e += model.x
        self.assertEqual(model.obj.expr(), 3.0)
        self.assertEqual(id(model.obj.expr.arg(1)), id(model.ec))

    def test_init_abstract_indexed(self):
        model = AbstractModel()
        model.ec = Expression([1, 2, 3], initialize=1.0)
        model.obj = Objective(rule=lambda m: 1.0 + sum_product(m.ec, index=[1, 2, 3]))
        inst = model.create_instance()
        self.assertEqual(inst.obj.expr(), 4.0)
        inst.ec[1].set_value(2.0)
        self.assertEqual(inst.obj.expr(), 5.0)

    def test_init_abstract_nonindexed(self):
        model = AbstractModel()
        model.y = Var(initialize=0.0)
        model.x = Var(initialize=1.0)
        model.ec = Expression(initialize=0.0)

        def obj_rule(model):
            return 1.0 + model.ec

        model.obj = Objective(rule=obj_rule)
        inst = model.create_instance()
        self.assertEqual(inst.obj.expr(), 1.0)
        self.assertEqual(id(inst.obj.expr.arg(1)), id(inst.ec))
        e = 1.0
        inst.ec.set_value(e)
        self.assertEqual(inst.obj.expr(), 2.0)
        self.assertEqual(id(inst.obj.expr.arg(1)), id(inst.ec))
        e += inst.x
        inst.ec.set_value(e)
        self.assertEqual(inst.obj.expr(), 3.0)
        self.assertEqual(id(inst.obj.expr.arg(1)), id(inst.ec))
        e += inst.x
        self.assertEqual(inst.obj.expr(), 3.0)
        self.assertEqual(id(inst.obj.expr.arg(1)), id(inst.ec))

        model.del_component('obj')
        model.del_component('ec')
        model.ec = Expression(initialize=0.0)

        def obj_rule(model):
            return 1.0 + model.ec

        model.obj = Objective(rule=obj_rule)
        inst = model.create_instance()
        self.assertEqual(inst.obj.expr(), 1.0)
        self.assertEqual(id(inst.obj.expr.arg(1)), id(inst.ec))
        e = 1.0
        inst.ec.set_value(e)
        self.assertEqual(inst.obj.expr(), 2.0)
        self.assertEqual(id(inst.obj.expr.arg(1)), id(inst.ec))
        e += inst.x
        inst.ec.set_value(e)
        self.assertEqual(inst.obj.expr(), 3.0)
        self.assertEqual(id(inst.obj.expr.arg(1)), id(inst.ec))
        e += inst.x
        self.assertEqual(inst.obj.expr(), 3.0)
        self.assertEqual(id(inst.obj.expr.arg(1)), id(inst.ec))

        model.del_component('obj')
        model.del_component('ec')
        model.ec = Expression(initialize=0.0)

        def obj_rule(model):
            return 1.0 + model.ec

        model.obj = Objective(rule=obj_rule)
        inst = model.create_instance()
        self.assertEqual(inst.obj.expr(), 1.0)
        self.assertEqual(id(inst.obj.expr.arg(1)), id(inst.ec))
        e = 1.0
        inst.ec.set_value(e)
        self.assertEqual(inst.obj.expr(), 2.0)
        self.assertEqual(id(inst.obj.expr.arg(1)), id(inst.ec))
        e += inst.x
        inst.ec.set_value(e)
        self.assertEqual(inst.obj.expr(), 3.0)
        self.assertEqual(id(inst.obj.expr.arg(1)), id(inst.ec))
        e += inst.x
        self.assertEqual(inst.obj.expr(), 3.0)
        self.assertEqual(id(inst.obj.expr.arg(1)), id(inst.ec))

    def test_pprint_oldStyle(self):
        expr_common.TO_STRING_VERBOSE = True

        model = ConcreteModel()
        model.x = Var()
        model.e = Expression(initialize=model.x + 2)
        model.E = Expression([1, 2], initialize=model.x**2 + 1)
        expr = model.e * model.x**2 + model.E[1]

        output = """\
sum(prod(e{sum(x, 2)}, pow(x, 2)), E[1]{sum(pow(x, 2), 1)})
e : Size=1, Index=None
    Key  : Expression
    None : sum(x, 2)
E : Size=2, Index={1, 2}
    Key : Expression
      1 : sum(pow(x, 2), 1)
      2 : sum(pow(x, 2), 1)
"""
        out = StringIO()
        out.write(str(expr) + "\n")
        model.e.pprint(ostream=out)
        # model.E[1].pprint(ostream=out)
        model.E.pprint(ostream=out)
        self.assertEqual(output, out.getvalue())

        model.e.set_value(1.0)
        model.E[1].set_value(2.0)
        output = """\
sum(prod(e{1.0}, pow(x, 2)), E[1]{2.0})
e : Size=1, Index=None
    Key  : Expression
    None :        1.0
E : Size=2, Index={1, 2}
    Key : Expression
      1 : 2.0
      2 : sum(pow(x, 2), 1)
"""
        out = StringIO()
        out.write(str(expr) + "\n")
        model.e.pprint(ostream=out)
        # model.E[1].pprint(ostream=out)
        model.E.pprint(ostream=out)
        self.assertEqual(output, out.getvalue())

        model.e.set_value(None)
        model.E[1].set_value(None)
        output = """\
sum(prod(e{Undefined}, pow(x, 2)), E[1]{Undefined})
e : Size=1, Index=None
    Key  : Expression
    None :  Undefined
E : Size=2, Index={1, 2}
    Key : Expression
      1 : Undefined
      2 : sum(pow(x, 2), 1)
"""
        out = StringIO()
        out.write(str(expr) + "\n")
        model.e.pprint(ostream=out)
        # model.E[1].pprint(ostream=out)
        model.E.pprint(ostream=out)
        self.assertEqual(output, out.getvalue())

    def test_pprint_newStyle(self):
        expr_common.TO_STRING_VERBOSE = False

        model = ConcreteModel()
        model.x = Var()
        model.e = Expression(initialize=model.x + 2)
        model.E = Expression([1, 2], initialize=model.x**2 + 1)
        expr = model.e * model.x**2 + model.E[1]

        output = """\
(x + 2)*x**2 + (x**2 + 1)
e : Size=1, Index=None
    Key  : Expression
    None : x + 2
E : Size=2, Index={1, 2}
    Key : Expression
      1 : x**2 + 1
      2 : x**2 + 1
"""
        out = StringIO()
        out.write(str(expr) + "\n")
        model.e.pprint(ostream=out)
        # model.E[1].pprint(ostream=out)
        model.E.pprint(ostream=out)
        self.assertEqual(output, out.getvalue())

        model.e.set_value(1.0)
        model.E[1].set_value(2.0)
        #
        # WEH - the 1.0 seems unnecessary here, but it results from
        # a fixed variable in a sub-expression.  I can't decide if this
        # is the expected behavior or not.
        #
        output = """\
x**2 + 2.0
e : Size=1, Index=None
    Key  : Expression
    None :        1.0
E : Size=2, Index={1, 2}
    Key : Expression
      1 : 2.0
      2 : x**2 + 1
"""
        out = StringIO()
        out.write(str(expr) + "\n")
        model.e.pprint(ostream=out)
        # model.E[1].pprint(ostream=out)
        model.E.pprint(ostream=out)
        self.assertEqual(output, out.getvalue())

        model.e.set_value(None)
        model.E[1].set_value(None)
        output = """\
e{None}*x**2 + E[1]{None}
e : Size=1, Index=None
    Key  : Expression
    None :  Undefined
E : Size=2, Index={1, 2}
    Key : Expression
      1 : Undefined
      2 : x**2 + 1
"""
        out = StringIO()
        out.write(str(expr) + "\n")
        model.e.pprint(ostream=out)
        # model.E[1].pprint(ostream=out)
        model.E.pprint(ostream=out)
        self.assertEqual(output, out.getvalue())

    def test_len(self):
        model = AbstractModel()
        model.e = Expression()

        self.assertEqual(len(model.e), 0)
        inst = model.create_instance()
        self.assertEqual(len(inst.e), 1)

    def test_None_key(self):
        model = AbstractModel()
        model.e = Expression()
        inst = model.create_instance()
        self.assertEqual(id(inst.e), id(inst.e[None]))

    def test_singleton_get_set(self):
        model = ConcreteModel()
        model.e = Expression()
        self.assertEqual(len(model.e), 1)
        self.assertEqual(model.e.expr, None)
        model.e.expr = 1
        self.assertEqual(model.e.expr, 1)
        model.e.expr += 2
        self.assertEqual(model.e.expr, 3)

    def test_singleton_get_set_value(self):
        model = ConcreteModel()
        model.e = Expression()
        self.assertEqual(len(model.e), 1)
        self.assertEqual(model.e.expr, None)
        model.e.expr = 1
        self.assertEqual(model.e.expr, 1)
        model.e.expr += 2
        self.assertEqual(model.e.expr, 3)

    def test_abstract_index(self):
        model = AbstractModel()
        model.A = Set()
        model.B = Set()
        model.C = model.A | model.B
        model.x = Expression(model.C)

    def test_iadd(self):
        # make sure simple for loops that look like they
        # create a new expression do not modify the named
        # expression
        m = ConcreteModel()
        e = m.e = Expression(expr=1.0)
        expr = 0.0
        for v in [1.0, e]:
            expr += v
        self.assertEqual(e.expr, 1)
        self.assertEqual(expr(), 2)
        expr = 0.0
        for v in [e, 1.0]:
            expr += v
        self.assertEqual(e.expr, 1)
        self.assertEqual(expr, 2)
        # Make sure that using in-place operators on named expressions
        # do not create loops inthe expression tree (test #1890)
        m.x = Var()
        m.y = Var()
        m.e.expr = m.x
        m.e += m.y
        assertExpressionsEqual(self, m.e.expr, m.x + m.y)

    def test_isub(self):
        # make sure simple for loops that look like they
        # create a new expression do not modify the named
        # expression
        m = ConcreteModel()
        e = m.e = Expression(expr=1.0)
        expr = 0.0
        for v in [1.0, e]:
            expr -= v
        self.assertEqual(e.expr, 1)
        self.assertEqual(expr(), -2)
        expr = 0.0
        for v in [e, 1.0]:
            expr -= v
        self.assertEqual(e.expr, 1)
        self.assertEqual(expr(), -2)
        # Make sure that using in-place operators on named expressions
        # do not create loops inthe expression tree (test #1890)
        m.x = Var()
        m.y = Var()
        m.e.expr = m.x
        m.e -= m.y
        assertExpressionsEqual(
            self,
            m.e.expr,
            EXPR.LinearExpression([m.x, EXPR.MonomialTermExpression((-1, m.y))]),
        )
        self.assertTrue(compare_expressions(m.e.expr, m.x - m.y))

    def test_imul(self):
        # make sure simple for loops that look like they
        # create a new expression do not modify the named
        # expression
        m = ConcreteModel()
        e = m.e = Expression(expr=3.0)
        expr = 1.0
        for v in [2.0, e]:
            expr *= v
        self.assertEqual(e.expr, 3)
        self.assertEqual(expr(), 6)
        expr = 1.0
        for v in [e, 2.0]:
            expr *= v
        self.assertEqual(e.expr, 3)
        self.assertEqual(expr, 6)
        # Make sure that using in-place operators on named expressions
        # do not create loops inthe expression tree (test #1890)
        m.x = Var()
        m.y = Var()
        m.e.expr = m.x
        m.e *= m.y
        self.assertTrue(compare_expressions(m.e.expr, m.x * m.y))

    def test_idiv(self):
        # make sure simple for loops that look like they
        # create a new expression do not modify the named
        # expression
        # floating point division
        m = ConcreteModel()
        e = m.e = Expression(expr=3.0)
        expr = e
        for v in [2.0, 1.0]:
            expr /= v
        self.assertEqual(e.expr, 3)
        self.assertEqual(expr, 1.5)
        expr = e
        for v in [1.0, 2.0]:
            expr /= v
        self.assertEqual(e.expr, 3)
        self.assertEqual(expr, 1.5)
        # note that integer division does not occur within
        # Pyomo expressions
        m = ConcreteModel()
        e = m.e = Expression(expr=3.0)
        expr = e
        for v in [2, 1]:
            expr /= v
        self.assertEqual(e.expr, 3)
        self.assertEqual(expr, 1.5)
        expr = e
        for v in [1, 2]:
            expr /= v
        self.assertEqual(e.expr, 3)
        self.assertEqual(expr, 1.5)
        # Make sure that using in-place operators on named expressions
        # do not create loops inthe expression tree (test #1890)
        m.x = Var()
        m.y = Var()
        m.e.expr = m.x
        m.e /= m.y
        self.assertTrue(compare_expressions(m.e.expr, m.x / m.y))

    def test_ipow(self):
        # make sure simple for loops that look like they
        # create a new expression do not modify the named
        # expression
        m = ConcreteModel()
        e = m.e = Expression(expr=3.0)
        expr = e
        for v in [2.0, 1.0]:
            expr **= v
        self.assertEqual(e.expr, 3)
        self.assertEqual(expr, 9)
        expr = e
        for v in [1.0, 2.0]:
            expr **= v
        self.assertEqual(e.expr, 3)
        self.assertEqual(expr, 9)
        # Make sure that using in-place operators on named expressions
        # do not create loops inthe expression tree (test #1890)
        m.x = Var()
        m.y = Var()
        m.e.expr = m.x
        m.e **= m.y
        self.assertTrue(compare_expressions(m.e.expr, m.x**m.y))


if __name__ == "__main__":
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
"""
Semibatch model, based on Nicholson et al. (2018). pyomo.dae: A modeling and
automatic discretization framework for optimization with di
erential and
algebraic equations. Mathematical Programming Computation, 10(2), 187-223.
"""

import json
from os.path import join, abspath, dirname
from pyomo.environ import (
    ConcreteModel,
    Set,
    Param,
    Var,
    Constraint,
    ConstraintList,
    Expression,
    Objective,
    TransformationFactory,
    SolverFactory,
    exp,
    minimize,
    Suffix,
    ComponentUID,
)
from pyomo.dae import ContinuousSet, DerivativeVar
from pyomo.contrib.parmest.experiment import Experiment


def generate_model(data):
    # if data is a file name, then load file first
    if isinstance(data, str):
        file_name = data
        try:
            with open(file_name, "r") as infile:
                data = json.load(infile)
        except:
            raise RuntimeError(f"Could not read {file_name} as json")

    # unpack and fix the data
    cameastemp = data["Ca_meas"]
    cbmeastemp = data["Cb_meas"]
    ccmeastemp = data["Cc_meas"]
    trmeastemp = data["Tr_meas"]

    cameas = {}
    cbmeas = {}
    ccmeas = {}
    trmeas = {}
    for i in cameastemp.keys():
        cameas[float(i)] = cameastemp[i]
        cbmeas[float(i)] = cbmeastemp[i]
        ccmeas[float(i)] = ccmeastemp[i]
        trmeas[float(i)] = trmeastemp[i]

    m = ConcreteModel()

    #
    # Measurement Data
    #
    m.measT = Set(initialize=sorted(cameas.keys()))
    m.Ca_meas = Param(m.measT, initialize=cameas)
    m.Cb_meas = Param(m.measT, initialize=cbmeas)
    m.Cc_meas = Param(m.measT, initialize=ccmeas)
    m.Tr_meas = Param(m.measT, initialize=trmeas)

    #
    # Parameters for semi-batch reactor model
    #
    m.R = Param(initialize=8.314)  # kJ/kmol/K
    m.Mwa = Param(initialize=50.0)  # kg/kmol
    m.rhor = Param(initialize=1000.0)  # kg/m^3
    m.cpr = Param(initialize=3.9)  # kJ/kg/K
    m.Tf = Param(initialize=300)  # K
    m.deltaH1 = Param(initialize=-40000.0)  # kJ/kmol
    m.deltaH2 = Param(initialize=-50000.0)  # kJ/kmol
    m.alphaj = Param(initialize=0.8)  # kJ/s/m^2/K
    m.alphac = Param(initialize=0.7)  # kJ/s/m^2/K
    m.Aj = Param(initialize=5.0)  # m^2
    m.Ac = Param(initialize=3.0)  # m^2
    m.Vj = Param(initialize=0.9)  # m^3
    m.Vc = Param(initialize=0.07)  # m^3
    m.rhow = Param(initialize=700.0)  # kg/m^3
    m.cpw = Param(initialize=3.1)  # kJ/kg/K
    m.Ca0 = Param(initialize=data["Ca0"])  # kmol/m^3)
    m.Cb0 = Param(initialize=data["Cb0"])  # kmol/m^3)
    m.Cc0 = Param(initialize=data["Cc0"])  # kmol/m^3)
    m.Tr0 = Param(initialize=300.0)  # K
    m.Vr0 = Param(initialize=1.0)  # m^3

    m.time = ContinuousSet(bounds=(0, 21600), initialize=m.measT)  # Time in seconds

    #
    # Control Inputs
    #
    def _initTc(m, t):
        if t < 10800:
            return data["Tc1"]
        else:
            return data["Tc2"]

    m.Tc = Param(
        m.time, initialize=_initTc, default=_initTc
    )  # bounds= (288,432) Cooling coil temp, control input

    def _initFa(m, t):
        if t < 10800:
            return data["Fa1"]
        else:
            return data["Fa2"]

    m.Fa = Param(
        m.time, initialize=_initFa, default=_initFa
    )  # bounds=(0,0.05) Inlet flow rate, control input

    #
    # Parameters being estimated
    #
    m.k1 = Var(initialize=14, bounds=(2, 100))  # 1/s Actual: 15.01
    m.k2 = Var(initialize=90, bounds=(2, 150))  # 1/s Actual: 85.01
    m.E1 = Var(initialize=27000.0, bounds=(25000, 40000))  # kJ/kmol Actual: 30000
    m.E2 = Var(initialize=45000.0, bounds=(35000, 50000))  # kJ/kmol Actual: 40000
    # m.E1.fix(30000)
    # m.E2.fix(40000)

    #
    # Time dependent variables
    #
    m.Ca = Var(m.time, initialize=m.Ca0, bounds=(0, 25))
    m.Cb = Var(m.time, initialize=m.Cb0, bounds=(0, 25))
    m.Cc = Var(m.time, initialize=m.Cc0, bounds=(0, 25))
    m.Vr = Var(m.time, initialize=m.Vr0)
    m.Tr = Var(m.time, initialize=m.Tr0)
    m.Tj = Var(
        m.time, initialize=310.0, bounds=(288, None)
    )  # Cooling jacket temp, follows coil temp until failure

    #
    # Derivatives in the model
    #
    m.dCa = DerivativeVar(m.Ca)
    m.dCb = DerivativeVar(m.Cb)
    m.dCc = DerivativeVar(m.Cc)
    m.dVr = DerivativeVar(m.Vr)
    m.dTr = DerivativeVar(m.Tr)

    #
    # Differential Equations in the model
    #

    def _dCacon(m, t):
        if t == 0:
            return Constraint.Skip
        return (
            m.dCa[t]
            == m.Fa[t] / m.Vr[t] - m.k1 * exp(-m.E1 / (m.R * m.Tr[t])) * m.Ca[t]
        )

    m.dCacon = Constraint(m.time, rule=_dCacon)

    def _dCbcon(m, t):
        if t == 0:
            return Constraint.Skip
        return (
            m.dCb[t]
            == m.k1 * exp(-m.E1 / (m.R * m.Tr[t])) * m.Ca[t]
            - m.k2 * exp(-m.E2 / (m.R * m.Tr[t])) * m.Cb[t]
        )

    m.dCbcon = Constraint(m.time, rule=_dCbcon)

    def _dCccon(m, t):
        if t == 0:
            return Constraint.Skip
        return m.dCc[t] == m.k2 * exp(-m.E2 / (m.R * m.Tr[t])) * m.Cb[t]

    m.dCccon = Constraint(m.time, rule=_dCccon)

    def _dVrcon(m, t):
        if t == 0:
            return Constraint.Skip
        return m.dVr[t] == m.Fa[t] * m.Mwa / m.rhor

    m.dVrcon = Constraint(m.time, rule=_dVrcon)

    def _dTrcon(m, t):
        if t == 0:
            return Constraint.Skip
        return m.rhor * m.cpr * m.dTr[t] == m.Fa[t] * m.Mwa * m.cpr / m.Vr[t] * (
            m.Tf - m.Tr[t]
        ) - m.k1 * exp(-m.E1 / (m.R * m.Tr[t])) * m.Ca[t] * m.deltaH1 - m.k2 * exp(
            -m.E2 / (m.R * m.Tr[t])
        ) * m.Cb[
            t
        ] * m.deltaH2 + m.alphaj * m.Aj / m.Vr0 * (
            m.Tj[t] - m.Tr[t]
        ) + m.alphac * m.Ac / m.Vr0 * (
            m.Tc[t] - m.Tr[t]
        )

    m.dTrcon = Constraint(m.time, rule=_dTrcon)

    def _singlecooling(m, t):
        return m.Tc[t] == m.Tj[t]

    m.singlecooling = Constraint(m.time, rule=_singlecooling)

    # Initial Conditions
    def _initcon(m):
        yield m.Ca[m.time.first()] == m.Ca0
        yield m.Cb[m.time.first()] == m.Cb0
        yield m.Cc[m.time.first()] == m.Cc0
        yield m.Vr[m.time.first()] == m.Vr0
        yield m.Tr[m.time.first()] == m.Tr0

    m.initcon = ConstraintList(rule=_initcon)

    #
    # Stage-specific cost computations
    #
    def ComputeFirstStageCost_rule(model):
        return 0

    m.FirstStageCost = Expression(rule=ComputeFirstStageCost_rule)

    def AllMeasurements(m):
        return sum(
            (m.Ca[t] - m.Ca_meas[t]) ** 2
            + (m.Cb[t] - m.Cb_meas[t]) ** 2
            + (m.Cc[t] - m.Cc_meas[t]) ** 2
            + 0.01 * (m.Tr[t] - m.Tr_meas[t]) ** 2
            for t in m.measT
        )

    def MissingMeasurements(m):
        if data["experiment"] == 1:
            return sum(
                (m.Ca[t] - m.Ca_meas[t]) ** 2
                + (m.Cb[t] - m.Cb_meas[t]) ** 2
                + (m.Cc[t] - m.Cc_meas[t]) ** 2
                + (m.Tr[t] - m.Tr_meas[t]) ** 2
                for t in m.measT
            )
        elif data["experiment"] == 2:
            return sum((m.Tr[t] - m.Tr_meas[t]) ** 2 for t in m.measT)
        else:
            return sum(
                (m.Cb[t] - m.Cb_meas[t]) ** 2 + (m.Tr[t] - m.Tr_meas[t]) ** 2
                for t in m.measT
            )

    m.SecondStageCost = Expression(rule=MissingMeasurements)

    def total_cost_rule(model):
        return model.FirstStageCost + model.SecondStageCost

    m.Total_Cost_Objective = Objective(rule=total_cost_rule, sense=minimize)

    # Discretize model
    disc = TransformationFactory("dae.collocation")
    disc.apply_to(m, nfe=20, ncp=4)
    return m


class SemiBatchExperiment(Experiment):

    def __init__(self, data):
        self.data = data
        self.model = None

    def create_model(self):
        self.model = generate_model(self.data)

    def label_model(self):

        m = self.model

        m.experiment_outputs = Suffix(direction=Suffix.LOCAL)
        m.experiment_outputs.update(
            (m.Ca[t], self.data["Ca_meas"][f"{t}"]) for t in m.measT
        )
        m.experiment_outputs.update(
            (m.Cb[t], self.data["Cb_meas"][f"{t}"]) for t in m.measT
        )
        m.experiment_outputs.update(
            (m.Cc[t], self.data["Cc_meas"][f"{t}"]) for t in m.measT
        )
        m.experiment_outputs.update(
            (m.Tr[t], self.data["Tr_meas"][f"{t}"]) for t in m.measT
        )

        m.unknown_parameters = Suffix(direction=Suffix.LOCAL)
        m.unknown_parameters.update(
            (k, ComponentUID(k)) for k in [m.k1, m.k2, m.E1, m.E2]
        )

    def finalize_model(self):
        pass

    def get_labeled_model(self):
        self.create_model()
        self.label_model()
        self.finalize_model()

        return self.model


def main():
    # Data loaded from files
    file_dirname = dirname(abspath(str(__file__)))
    file_name = abspath(join(file_dirname, "exp2.out"))
    with open(file_name, "r") as infile:
        data = json.load(infile)
    data["experiment"] = 2

    model = generate_model(data)
    solver = SolverFactory("ipopt")
    solver.solve(model)
    print("k1 = ", model.k1())
    print("E1 = ", model.E1())


if __name__ == "__main__":
    main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.common.dependencies import attempt_import
from pyomo.common.numeric_types import native_types
from pyomo.common.modeling import NOTSET
from pyomo.core.pyomoobject import PyomoObject
from pyomo.core.expr.expr_common import OperatorAssociativity, _type_check_exception_arg

visitor, _ = attempt_import('pyomo.core.expr.visitor')


class ExpressionBase(PyomoObject):
    """The base class for all Pyomo expression systems.

    This class is used to define nodes in a general expression tree.
    Individual expression systems (numeric, logical, etc.) will mix this
    class in with their fundamental base data type (NumericValue,
    BooleanValue, etc) to form the base node of that expression system.
    """

    __slots__ = ()

    PRECEDENCE = 0

    # Most operators in Python are left-to-right associative
    """Return the associativity of this operator.

    Returns 1 if this operator is left-to-right associative or -1 if
    it is right-to-left associative.  Any other return value will be
    interpreted as "not associative" (implying any arguments that
    are at this operator's PRECEDENCE will be enclosed in parens).
    """
    ASSOCIATIVITY = OperatorAssociativity.LEFT_TO_RIGHT

    def nargs(self):
        """Returns the number of child nodes.

        Note
        ----

        Individual expression nodes may use different internal storage
        schemes, so it is imperative that developers use this method and
        not assume the existence of a particular attribute!

        Returns
        -------
        int: A nonnegative integer that is the number of child nodes.

        """
        raise NotImplementedError(
            f"Derived expression ({self.__class__}) failed to implement nargs()"
        )

    def arg(self, i):
        """Return the i-th child node.

        Parameters
        ----------
        i: int
            Index of the child argument to return

        Returns: The i-th child node.

        """
        if i < 0:
            i += self.nargs()
            if i < 0:
                raise KeyError(
                    "Invalid index for expression argument: %d" % i - self.nargs()
                )
        elif i >= self.nargs():
            raise KeyError("Invalid index for expression argument: %d" % i)
        return self._args_[i]

    @property
    def args(self):
        """Return the child nodes

        Returns
        -------
        list or tuple:
            Sequence containing only the child nodes of this node.  The
            return type depends on the node storage model.  Users are
            not permitted to change the returned data (even for the case
            of data returned as a list), as that breaks the promise of
            tree immutability.

        """
        raise NotImplementedError(
            f"Derived expression ({self.__class__}) failed to implement args()"
        )

    def __call__(self, exception=NOTSET):
        """Evaluate the value of the expression tree.

        Parameters
        ----------
        exception: bool
            If :const:`False`, then an exception raised while evaluating
            is captured, and the value returned is :const:`None`.
            Default is :const:`True`.

        Returns
        -------
        The value of the expression or :const:`None`.

        """
        exception = _type_check_exception_arg(self, exception)
        return visitor.evaluate_expression(self, exception)

    def __str__(self):
        """Returns a string description of the expression.

        Note:

        The value of ``pyomo.core.expr.expr_common.TO_STRING_VERBOSE``
        is used to configure the execution of this method.  If this
        value is :const:`True`, then the string representation is a
        nested function description of the expression.  The default is
        :const:`False`, which returns an algebraic (infix notation)
        description of the expression.

        Returns
        -------
        str
        """
        return visitor.expression_to_string(self)

    def to_string(self, verbose=None, labeler=None, smap=None, compute_values=False):
        """Return a string representation of the expression tree.

        Parameters
        ----------
        verbose: bool
            If :const:`True`, then the string representation
            consists of nested functions.  Otherwise, the string
            representation is an algebraic (infix notation) equation.
            Defaults to :const:`False`.

        labeler:
            An object that generates string labels for variables in the
            expression tree.  Defaults to :const:`None`.

        smap:
            If specified, this
            :class:`SymbolMap <pyomo.core.expr.symbol_map.SymbolMap>`
            is used to cache labels for variables.

        compute_values (bool):
            If :const:`True`, then parameters and fixed variables are
            evaluated before the expression string is generated.
            Default is :const:`False`.

        Returns:
            A string representation for the expression tree.

        """
        return visitor.expression_to_string(
            self,
            verbose=verbose,
            labeler=labeler,
            smap=smap,
            compute_values=compute_values,
        )

    def _to_string(self, values, verbose, smap):
        """
        Construct a string representation for this node, using the string
        representations of its children.

        This method is called by the :class:`_ToStringVisitor
        <pyomo.core.expr.current._ToStringVisitor>` class.  It must
        must be defined in subclasses.

        Args:
            values (list): The string representations of the children of this
                node.
            verbose (bool): If :const:`True`, then the string
                representation consists of nested functions.  Otherwise,
                the string representation is an algebraic equation.
            smap:  If specified, this :class:`SymbolMap
                <pyomo.core.expr.symbol_map.SymbolMap>` is
                used to cache labels for variables.

        Returns:
            A string representation for this node.
        """
        raise NotImplementedError(
            f"Derived expression ({self.__class__}) failed to implement _to_string()"
        )

    def getname(self, *args, **kwds):
        """Return the text name of a function associated with this expression
        object.

        In general, no arguments are passed to this function.

        Args:
            *arg: a variable length list of arguments
            **kwds: keyword arguments

        Returns:
            A string name for the function.

        """
        raise NotImplementedError(
            f"Derived expression ({self.__class__}) failed to implement getname()"
        )

    def clone(self, substitute=None):
        """
        Return a clone of the expression tree.

        Note:
            This method does not clone the leaves of the
            tree, which are numeric constants and variables.
            It only clones the interior nodes, and
            expression leaf nodes like
            :class:`_MutableLinearExpression<pyomo.core.expr.current._MutableLinearExpression>`.
            However, named expressions are treated like
            leaves, and they are not cloned.

        Args:
            substitute (dict): a dictionary that maps object ids to clone
                objects generated earlier during the cloning process.

        Returns:
            A new expression tree.
        """
        return visitor.clone_expression(self, substitute=substitute)

    def create_node_with_local_data(self, args, classtype=None):
        """
        Construct a node using given arguments.

        This method provides a consistent interface for constructing a
        node, which is used in tree visitor scripts.  In the simplest
        case, this returns::

            self.__class__(args)

        But in general this creates an expression object using local
        data as well as arguments that represent the child nodes.

        Args:
            args (list): A list of child nodes for the new expression
                object

        Returns:
            A new expression object with the same type as the current
            class.
        """
        if classtype is None:
            classtype = self.__class__
        return classtype(args)

    def is_constant(self):
        """Return True if this expression is an atomic constant

        This method contrasts with the is_fixed() method.  This method
        returns True if the expression is an atomic constant, that is it
        is composed exclusively of constants and immutable parameters.
        NumericValue objects returning is_constant() == True may be
        simplified to their numeric value at any point without warning.

        Note:  This defaults to False, but gets redefined in sub-classes.
        """
        return False

    def is_fixed(self):
        """
        Return :const:`True` if this expression contains no free variables.

        Returns:
            A boolean.
        """
        return visitor._expression_is_fixed(self)

    def _is_fixed(self, values):
        """
        Compute whether this expression is fixed given
        the fixed values of its children.

        This method is called by the :class:`_IsFixedVisitor
        <pyomo.core.expr.current._IsFixedVisitor>` class.  It can
        be over-written by expression classes to customize this
        logic.

        Args:
            values (list): A list of boolean values that indicate whether
                the children of this expression are fixed

        Returns:
            A boolean that is :const:`True` if the fixed values of the
            children are all :const:`True`.
        """
        return all(values)

    def is_potentially_variable(self):
        """
        Return :const:`True` if this expression might represent
        a variable expression.

        This method returns :const:`True` when (a) the expression
        tree contains one or more variables, or (b) the expression
        tree contains a named expression. In both cases, the
        expression cannot be treated as constant since (a) the variables
        may not be fixed, or (b) the named expressions may be changed
        at a later time to include non-fixed variables.

        Returns:
            A boolean.  Defaults to :const:`True` for expressions.
        """
        return True

    def is_named_expression_type(self):
        """
        Return :const:`True` if this object is a named expression.

        This method returns :const:`False` for this class, and it
        is included in other classes within Pyomo that are not named
        expressions, which allows for a check for named expressions
        without evaluating the class type.

        Returns:
            A boolean.
        """
        return False

    def is_expression_type(self, expression_system=None):
        """
        Return :const:`True` if this object is an expression.

        This method obviously returns :const:`True` for this class, but it
        is included in other classes within Pyomo that are not expressions,
        which allows for a check for expressions without
        evaluating the class type.

        Returns:
            A boolean.
        """
        return expression_system is None or expression_system == self.EXPRESSION_SYSTEM

    def size(self):
        """
        Return the number of nodes in the expression tree.

        Returns:
            A nonnegative integer that is the number of interior and leaf
            nodes in the expression tree.
        """
        return visitor.sizeof_expression(self)

    def _apply_operation(self, result):
        """
        Compute the values of this node given the values of its children.

        This method is called by the :class:`_EvaluationVisitor
        <pyomo.core.expr.current._EvaluationVisitor>` class.  It must
        be over-written by expression classes to customize this logic.

        Note:
            This method applies the logical operation of the
            operator to the arguments.  It does *not* evaluate
            the arguments in the process, but assumes that they
            have been previously evaluated.  But note that if
            this class contains auxiliary data (e.g. like the
            numeric coefficients in the :class:`LinearExpression
            <pyomo.core.expr.current.LinearExpression>` class) then
            those values *must* be evaluated as part of this
            function call.  An uninitialized parameter value
            encountered during the execution of this method is
            considered an error.

        Args:
            values (list): A list of values that indicate the value
                of the children expressions.

        Returns:
            A floating point value for this expression.
        """
        raise NotImplementedError(
            f"Derived expression ({self.__class__}) failed to "
            "implement _apply_operation()"
        )


class NPV_Mixin:
    __slots__ = ()

    def is_potentially_variable(self):
        return False

    def create_node_with_local_data(self, args, classtype=None):
        assert classtype is None
        try:
            npv_args = all(
                type(arg) in native_types or not arg.is_potentially_variable()
                for arg in args
            )
        except AttributeError:
            # We can hit this during expression replacement when the new
            # type is not a PyomoObject type, but is not in the
            # native_types set.  We will play it safe and clear the NPV flag
            npv_args = False
        if npv_args:
            return super().create_node_with_local_data(args, None)
        else:
            return super().create_node_with_local_data(
                args, self.potentially_variable_base_class()
            )

    def potentially_variable_base_class(self):
        cls = list(self.__class__.__bases__)
        cls.remove(NPV_Mixin)
        assert len(cls) == 1
        return cls[0]


class ExpressionArgs_Mixin:
    __slots__ = ('_args_',)

    def __init__(self, args):
        self._args_ = args

    def nargs(self):
        return len(self._args_)

    @property
    def args(self):
        """
        Return the child nodes

        Returns
        -------
        list or tuple:
            Sequence containing only the child nodes of this node.  The
            return type depends on the node storage model.  Users are
            not permitted to change the returned data (even for the case
            of data returned as a list), as that breaks the promise of
            tree immutability.
        """
        return self._args_



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
#
# This module was originally developed as part of the IDAES PSE Framework
#
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

"""
Test data model items for QTreeView. These tests should work even without PyQt.
"""

import pyomo.common.unittest as unittest
from pyomo.common.collections import ComponentSet
from pyomo.environ import (
    ConcreteModel,
    Var,
    BooleanVar,
    Block,
    Param,
    Expression,
    Constraint,
    Objective,
    ExternalFunction,
    Reals,
    log,
    sin,
    sqrt,
    expr,
)
import pyomo.contrib.viewer.report as rpt
import pyomo.contrib.viewer.ui_data as uidata


class TestReportFunctions(unittest.TestCase):
    def setUp(self):
        # Borrowed this test model from the trust region tests
        m = ConcreteModel()
        m.z = Var(range(3), domain=Reals, initialize=2.0)
        m.x = Var(range(4), initialize=2.0)
        m.x[1] = 1.0
        m.x[2] = 0.0
        m.x[3] = None

        m.b1 = Block()
        m.b1.e1 = Expression(expr=m.x[0] + m.x[1])
        m.b1.e2 = Expression(expr=m.x[0] / m.x[2])
        m.b1.e3 = Expression(expr=m.x[3] * m.x[1])
        m.b1.e4 = Expression(expr=log(m.x[2]))
        m.b1.e5 = Expression(expr=log(m.x[2] - 2))

        def blackbox(a, b):
            return sin(a - b)

        self.bb = ExternalFunction(blackbox)

        m.obj = Objective(
            expr=(m.z[0] - 1.0) ** 2
            + (m.z[0] - m.z[1]) ** 2
            + (m.z[2] - 1.0) ** 2
            + (m.x[0] - 1.0) ** 4
            + (m.x[1] - 1.0) ** 6  # + m.bb(m.x[0],m.x[1])
        )
        m.c1 = Constraint(
            expr=m.x[0] * m.z[0] ** 2 + self.bb(m.x[0], m.x[1]) == 2 * sqrt(2.0)
        )
        m.c2 = Constraint(expr=m.z[2] ** 4 * m.z[1] ** 2 + m.z[1] == 8 + sqrt(2.0))
        m.c3 = Constraint(expr=m.x[1] == 3)
        m.c4 = Constraint(expr=0 == 3 / m.x[2])
        m.c5 = Constraint(expr=0 == log(m.x[2]))
        m.c6 = Constraint(expr=0 == log(m.x[2] - 4))
        m.c7 = Constraint(expr=0 == log(m.x[3]))
        m.p1 = Param(mutable=True, initialize=1)
        m.c8 = Constraint(expr=m.x[1] <= 1 / m.p1)
        m.c8b = Constraint(expr=m.x[1] >= 1 / m.p1)
        m.c9 = Constraint(expr=m.x[1] <= 1)
        m.c10 = Constraint(expr=m.x[1] >= 1)
        m.p1 = 0
        self.m = m.clone()

    def test_value_no_exception(self):
        # Try to divide by zero
        self.m.x[2] = 0
        v = rpt.value_no_exception(self.m.b1.e2, div0="I like to divide by zero")
        assert v == "I like to divide by zero"
        # Try as calculation with None
        self.m.x[2] = None
        v = rpt.value_no_exception(self.m.b1.e2, div0=None)
        assert v is None
        # Try log of negative number
        self.m.x[2] = 0.0
        v = rpt.value_no_exception(self.m.b1.e5)
        assert v is None
        # Try a valid calculation
        self.m.x[2] = 2.0
        v = rpt.value_no_exception(self.m.b1.e2, div0=None)
        self.assertAlmostEqual(v, 1)

    def test_get_residual(self):
        dat = uidata.UIData(self.m)
        # so that the model viewer doesn't run slow on large models,
        # you have to explicitly ask for constraints and expressions
        # to be calculated. Getting the residual before calculation
        # should just give None
        assert rpt.get_residual(dat, self.m.c3) is None
        dat.calculate_constraints()
        self.assertAlmostEqual(rpt.get_residual(dat, self.m.c3), 2.0)
        # In c8 the bound has a divide by 0, I think this is only possible
        # with a mutable param
        assert rpt.get_residual(dat, self.m.c8) == "Divide_by_0"
        assert rpt.get_residual(dat, self.m.c8b) == "Divide_by_0"
        self.m.x[2] = 0
        assert rpt.get_residual(dat, self.m.c4) == "Divide_by_0"
        self.m.x[2] = 2
        # haven't recalculated so still error
        assert rpt.get_residual(dat, self.m.c4) == "Divide_by_0"
        dat.calculate_constraints()
        self.assertAlmostEqual(rpt.get_residual(dat, self.m.c4), 3.0 / 2.0)
        self.assertAlmostEqual(rpt.get_residual(dat, self.m.c9), 0)
        self.assertAlmostEqual(rpt.get_residual(dat, self.m.c10), 0)

    def test_active_equalities(self):
        eq = [
            self.m.c1,
            self.m.c2,
            self.m.c3,
            self.m.c4,
            self.m.c5,
            self.m.c6,
            self.m.c7,
        ]
        for i, o in enumerate(rpt.active_equalities(self.m)):
            assert o == eq[i]

    def test_active_constraint_set(self):
        self.m.c4.deactivate()
        assert rpt.active_constraint_set(self.m) == ComponentSet(
            [
                self.m.c1,
                self.m.c2,
                self.m.c3,
                self.m.c5,
                self.m.c6,
                self.m.c7,
                self.m.c8,
                self.m.c8b,
                self.m.c9,
                self.m.c10,
            ]
        )
        self.m.c4.activate()
        assert rpt.active_constraint_set(self.m) == ComponentSet(
            [
                self.m.c1,
                self.m.c2,
                self.m.c3,
                self.m.c4,
                self.m.c5,
                self.m.c6,
                self.m.c7,
                self.m.c8,
                self.m.c8b,
                self.m.c9,
                self.m.c10,
            ]
        )

    def test_active_equality_set(self):
        self.m.c4.deactivate()
        assert rpt.active_equality_set(self.m) == ComponentSet(
            [self.m.c1, self.m.c2, self.m.c3, self.m.c5, self.m.c6, self.m.c7]
        )
        self.m.c4.activate()
        assert rpt.active_equality_set(self.m) == ComponentSet(
            [
                self.m.c1,
                self.m.c2,
                self.m.c3,
                self.m.c4,
                self.m.c5,
                self.m.c6,
                self.m.c7,
            ]
        )

    def test_count_free_variables(self):
        assert rpt.count_free_variables(self.m) == 7

    def test_count_equality_constraints(self):
        assert rpt.count_equality_constraints(self.m) == 7

    def test_count_constraints(self):
        assert rpt.count_constraints(self.m) == 11

    def test_degrees_of_freedom(self):
        assert rpt.degrees_of_freedom(self.m) == 0



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
#
# This module was originally developed as part of the IDAES PSE Framework
#
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

"""
A simple GUI viewer/editor for Pyomo models.
"""

__author__ = "John Eslick"

import os
import logging

try:
    from IPython import get_ipython
except ImportError:

    def get_ipython():
        raise AttributeError("IPython not available")


import pyomo.contrib.viewer.report as rpt
import pyomo.environ as pyo
import pyomo.contrib.viewer.qt as myqt

from pyomo.common.fileutils import this_file_dir
from pyomo.common.flags import building_documentation
from pyomo.contrib.viewer.model_browser import ModelBrowser
from pyomo.contrib.viewer.residual_table import ResidualTable
from pyomo.contrib.viewer.model_select import ModelSelect
from pyomo.contrib.viewer.ui_data import UIData

_log = logging.getLogger(__name__)


# This lets the file be imported when the Qt UI is not available (or
# when building docs), but you won't be able to use it.  Allowing this
# will let some basic tests run (and pass) without PyQt
class _MainWindowUI:
    pass


class _MainWindow:
    pass


# Note that the classes loaded here have signatures that are not
# parsable by Sphinx, so we won't attempt to import them if we are
# building the API documentation.
if not building_documentation():
    _mypath = this_file_dir()
    try:
        _MainWindowUI, _MainWindow = myqt.uic.loadUiType(
            os.path.join(_mypath, "main.ui")
        )
    except:
        _log.exception("Failed to load UI files.")

for _err in myqt.import_errors:
    _log.error(_err)


def get_mainwindow(
    model=None, show=True, ask_close=True, model_var_name_in_main=None, testing=False
):
    """
    Create a UI MainWindow.

    Args:
        model: A Pyomo model to work with
        show: show the window after it is created
        ask_close: confirm close window
        testing: if True, expect testing
    Returns:
        (ui, model): ui is the MainWindow widget, and model is the linked Pyomo
            model.  If no model is provided a new ConcreteModel is created
    """
    model_name = model_var_name_in_main
    if model is None:
        import __main__

        if model_name in dir(__main__):
            if isinstance(getattr(__main__, model_name), pyo.Block):
                model = getattr(__main__, model_name)
        else:
            for s in dir(__main__):
                if isinstance(getattr(__main__, s), pyo.Block):
                    model = getattr(__main__, s)
                    model_name = s
                    break
    ui = MainWindow(
        model=model,
        model_var_name_in_main=model_name,
        ask_close=ask_close,
        testing=testing,
    )
    try:
        get_ipython().events.register("post_execute", ui.refresh_on_execute)
    except AttributeError:
        pass  # not in ipy kernel, so is fine to not register callback
    if show:
        ui.show()
    return ui


class MainWindow(_MainWindow, _MainWindowUI):
    def __init__(self, *args, **kwargs):
        model = self.model = kwargs.pop("model", None)
        main = self.main = kwargs.pop("main", None)
        ask_close = self.ask_close = kwargs.pop("ask_close", True)
        self.testing = kwargs.pop("testing", False)
        model_var_name_in_main = kwargs.pop("model_var_name_in_main", None)
        flags = kwargs.pop("flags", 0)
        self.ui_data = UIData(model=model)
        super().__init__(*args, **kwargs)
        self.setupUi(self)
        self.setCentralWidget(self.mdiArea)

        self._refresh_list = []
        self.variables = None
        self.constraints = None
        self.expressions = None
        self.parameters = None
        self.residuals = None
        self.update_model()

        self.ui_data.updated.connect(self.update_model)
        # Set menu actions (remember the menu items are defined in the ui file)
        # you can edit the menus in qt-designer
        self.actionModel_Selector.triggered.connect(self.show_model_select)
        self.ui_data.exec_refresh.connect(self.refresh_on_execute)
        self.actionRestart_Variable_View.triggered.connect(self.variables_restart)
        self.actionRestart_Constraint_View.triggered.connect(self.constraints_restart)
        self.actionRestart_Parameter_View.triggered.connect(self.parameters_restart)
        self.actionRestart_Expression_View.triggered.connect(self.expressions_restart)
        self.actionRestart_Residual_Table.triggered.connect(self.residuals_restart)
        self.actionInformation.triggered.connect(self.model_information)
        self.actionCalculateConstraints.triggered.connect(
            self.ui_data.calculate_constraints
        )
        self.actionCalculateExpressions.triggered.connect(
            self.ui_data.calculate_expressions
        )
        self.ui_data.model_var_name_in_main = model_var_name_in_main
        self.actionTile.triggered.connect(self.mdiArea.tileSubWindows)
        self.actionCascade.triggered.connect(self.mdiArea.cascadeSubWindows)
        self.actionTabs.triggered.connect(self.toggle_tabs)
        self._dialog = None  # dialog displayed so can access it easier for tests
        self._dialog_test_button = None  # button clicked on dialog in test mode
        self.mdiArea.setViewMode(myqt.QMdiArea.ViewMode.TabbedView)

    def toggle_tabs(self):
        if self.mdiArea.viewMode() == myqt.QMdiArea.ViewMode.SubWindowView:
            self.mdiArea.setViewMode(myqt.QMdiArea.ViewMode.TabbedView)
        elif self.mdiArea.viewMode() == myqt.QMdiArea.ViewMode.TabbedView:
            self.mdiArea.setViewMode(myqt.QMdiArea.ViewMode.SubWindowView)
        else:
            # There are no other modes unless there is a change in Qt so pass
            pass

    def _tree_restart(self, w, cls=ModelBrowser, **kwargs):
        """
        Start/Restart a tree window
        """
        try:
            self._refresh_list.remove(w)
        except ValueError:  # not in list? that's okay
            pass
        try:
            try:
                self.mdiArea.removeSubWindow(w.parent())
            except RuntimeError:  # user closed with "X" button
                pass
            del w
            w = None
        except AttributeError:
            pass
        w = cls(**kwargs)
        self.mdiArea.addSubWindow(w)
        w.parent().show()  # parent is now a MdiAreaSubWindow
        self._refresh_list.append(w)
        return w

    def variables_restart(self):
        self.variables = self._tree_restart(
            w=self.variables, standard="Var", ui_data=self.ui_data
        )

    def expressions_restart(self):
        self.expressions = self._tree_restart(
            w=self.expressions, standard="Expression", ui_data=self.ui_data
        )

    def parameters_restart(self):
        self.parameters = self._tree_restart(
            w=self.parameters, standard="Param", ui_data=self.ui_data
        )

    def constraints_restart(self):
        self.constraints = self._tree_restart(
            w=self.constraints, standard="Constraint", ui_data=self.ui_data
        )

    def residuals_restart(self):
        self.residuals = self._tree_restart(
            w=self.residuals, cls=ResidualTable, ui_data=self.ui_data
        )

    def set_model(self, model):
        self.ui_data.model = model

    def update_model(self):
        """
        Play it safe by restarting all the tree view widgets when the model updates
        """
        self.variables_restart()
        self.expressions_restart()
        self.constraints_restart()
        self.parameters_restart()
        self.mdiArea.setActiveSubWindow(self.variables.parent())
        self.toggle_tabs()
        self.toggle_tabs()

    def model_information(self):
        """
        Put some useful model information into a message box

        Displays:
        * number of active equality constraints
        * number of free variables in equality constraints
        * degrees of freedom

        Other things that could be added
        * number of deactivated equalities
        * number of active inequality constraints
        * number of deactivated inequality constraints
        * number of free variables not appearing in active constraints
        * number of fixed variables not appearing in active constraints
        * number of free variables not appearing in any constraints
        * number of fixed variables not appearing in any constraints
        * number of fixed variables appearing in constraints
        """
        active_eq = rpt.count_equality_constraints(self.ui_data.model)
        free_vars = rpt.count_free_variables(self.ui_data.model)
        cons = rpt.count_constraints(self.ui_data.model)
        dof = free_vars - active_eq
        if dof == 1:
            doftext = "Degree"
        else:
            doftext = "Degrees"
        msg = myqt.QMessageBox()
        msg.setStyleSheet("QLabel{min-width: 600px;}")
        self._dialog = msg
        msg.setWindowTitle("Model Information")
        msg.setText("""{} -- Active Constraints
{} -- Active Equalities
{} -- Free Variables
{} -- {} of Freedom""".format(cons, active_eq, free_vars, dof, doftext))
        msg.setStandardButtons(myqt.QMessageBox.StandardButton.Ok)
        msg.setModal(False)
        msg.show()

    def refresh_on_execute(self):
        """
        This is the call back function that happens when code is executed in the
        ipython kernel.  The main purpose of this right now it to refresh the
        UI display so that it matches the current state of the model.
        """
        if self.ui_data.model_var_name_in_main is not None:
            import __main__

            try:
                mname = self.ui_data.model_var_name_in_main
                mid = id(getattr(__main__, mname))
                if id(self.ui_data.model) != mid:
                    self.ui_data.model = getattr(__main__, mname)
                    self.update_model
                    return
            except AttributeError:
                pass
        for w in self._refresh_list:
            try:
                w.refresh()
            except RuntimeError:  # window closed by user pushing "X" button
                pass

    def show_model_select(self):
        model_select = ModelSelect(parent=self, ui_data=self.ui_data)
        model_select.update_models()
        model_select.show()
        return model_select

    def exit_action(self):
        """
        Selecting exit from the UI, triggers the close event on this mainwindow
        """
        self.close()

    def closeEvent(self, event):
        """
        Handle the close event by asking for confirmation
        """
        if not self.ask_close:
            event.accept()
            return
        msg = myqt.QMessageBox()
        self._dialog = msg
        msg.setIcon(myqt.QMessageBox.Icon.Question)
        msg.setText(
            "Are you sure you want to close this window?"
            " You can reopen it with ui.show()."
        )
        msg.setWindowTitle("Close?")
        msg.setStandardButtons(
            myqt.QMessageBox.StandardButton.Yes | myqt.QMessageBox.StandardButton.No
        )
        if self.testing:  # don't even show dialog just pretend button clicked
            result = self._dialog_test_button
        else:
            result = msg.exec()
        if result == myqt.QMessageBox.StandardButton.Yes:
            event.accept()
        else:
            event.ignore()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import logging

logger = logging.getLogger('pyomo.core')

_virtual_sets = []


#
# Dummy types used by Kernel as domain flags
#
class RealSet:
    @staticmethod
    def get_interval():
        return (None, None, 0)

    @staticmethod
    def is_continuous():
        return True

    @staticmethod
    def is_integer():
        return False

    @staticmethod
    def is_binary():
        return False


class IntegerSet:
    @staticmethod
    def get_interval():
        return (None, None, 1)

    @staticmethod
    def is_continuous():
        return False

    @staticmethod
    def is_integer():
        return True

    @staticmethod
    def is_binary():
        return False


class BinarySet:
    @staticmethod
    def get_interval():
        return (0, 1, 1)

    @staticmethod
    def is_continuous():
        return False

    @staticmethod
    def is_integer():
        return True

    @staticmethod
    def is_binary():
        return True


# TODO: Deprecate BooleanSet (that will soon be replaced by a true BooleanSet
#      admitting {True, False})
BooleanSet = BinarySet



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
#
# Author:  Gabe Hackebeil
# Purpose: For testing to ensure that the Pyomo NL writer properly
#          converts the nonlinear expression to the NL file format.
#
#          This test model relies on the gjh_asl_json executable. It
#          will not solve if sent to a real optimizer.
#

from pyomo.environ import ConcreteModel, Var, Objective, Constraint, maximize

model = ConcreteModel()

model.x = Var(initialize=0.5)

model.obj = Objective(expr=model.x, sense=maximize)

model.c1 = Constraint(expr=(model.x**3 - model.x) == 0)
model.c2 = Constraint(expr=10 * (model.x**3 - model.x) == 0)
model.c3 = Constraint(expr=(model.x**3 - model.x) / 10.0 == 0)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

# TODO: this import is for historical backwards compatibility and should
# probably be removed
from pyomo.common.collections import ComponentMap
from pyomo.common.enums import minimize, maximize

from pyomo.core.expr.symbol_map import SymbolMap
from pyomo.core.expr.numvalue import (
    nonpyomo_leaf_types,
    native_types,
    native_numeric_types,
    value,
    is_constant,
    is_fixed,
    is_variable_type,
    is_potentially_variable,
    polynomial_degree,
    NumericValue,
    ZeroConstant,
)
from pyomo.core.expr.boolean_value import (
    as_boolean,
    BooleanConstant,
    BooleanValue,
    native_logical_values,
)

from pyomo.core.base.component import name, Component, ModelComponentFactory
from pyomo.core.base.componentuid import ComponentUID
from pyomo.core.base.config import PyomoOptions
from pyomo.core.base.enums import SortComponents, TraversalStrategy, VarCollector
from pyomo.core.base.label import (
    CuidLabeler,
    CounterLabeler,
    NumericLabeler,
    CNameLabeler,
    TextLabeler,
    AlphaNumericTextLabeler,
    NameLabeler,
    ShortNameLabeler,
)
from pyomo.core.base.misc import display
from pyomo.core.base.reference import Reference
from pyomo.core.base.symbol_map import symbol_map_from_instance
from pyomo.core.base.transformation import (
    Transformation,
    TransformationFactory,
    ReverseTransformationToken,
)

from pyomo.core.base.PyomoModel import (
    global_option,
    ModelSolution,
    ModelSolutions,
    Model,
    ConcreteModel,
    AbstractModel,
)

#
# Components
#
from pyomo.core.base.action import BuildAction
from pyomo.core.base.block import (
    Block,
    BlockData,
    ScalarBlock,
    active_components,
    components,
    active_components_data,
    components_data,
)
from pyomo.core.base.boolean_var import (
    BooleanVar,
    BooleanVarData,
    BooleanVarList,
    ScalarBooleanVar,
)
from pyomo.core.base.check import BuildCheck
from pyomo.core.base.connector import Connector, ConnectorData
from pyomo.core.base.constraint import (
    simple_constraint_rule,
    simple_constraintlist_rule,
    ConstraintList,
    Constraint,
    ConstraintData,
)
from pyomo.core.base.expression import Expression, NamedExpressionData, ExpressionData
from pyomo.core.base.external import ExternalFunction
from pyomo.core.base.logical_constraint import (
    LogicalConstraint,
    LogicalConstraintList,
    LogicalConstraintData,
)
from pyomo.core.base.objective import (
    simple_objective_rule,
    simple_objectivelist_rule,
    Objective,
    ObjectiveList,
    ObjectiveData,
)
from pyomo.core.base.param import Param, ParamData
from pyomo.core.base.piecewise import Piecewise, PiecewiseData
from pyomo.core.base.set import (
    Set,
    SetData,
    SetOf,
    RangeSet,
    Reals,
    PositiveReals,
    NonPositiveReals,
    NegativeReals,
    NonNegativeReals,
    Integers,
    PositiveIntegers,
    NonPositiveIntegers,
    NegativeIntegers,
    NonNegativeIntegers,
    Boolean,
    Binary,
    Any,
    AnyWithNone,
    EmptySet,
    UnitInterval,
    PercentFraction,
    RealInterval,
    IntegerInterval,
    simple_set_rule,
)
from pyomo.core.base.sos import SOSConstraint, SOSConstraintData
from pyomo.core.base.suffix import (
    active_export_suffix_generator,
    active_import_suffix_generator,
    Suffix,
)
from pyomo.core.base.var import Var, VarData, ScalarVar, VarList

from pyomo.core.base.instance2dat import instance2dat

#
# These APIs are deprecated and should be removed in the near future
#
from pyomo.core.base.set import set_options, RealSet, IntegerSet, BooleanSet

#
# declare deprecation paths for removed modules and attributes
#
from pyomo.common.deprecation import relocated_module_attribute, moved_module

moved_module(
    "pyomo.core.base.plugin",
    "pyomo._archive.plugin",
    msg="The pyomo.core.base.plugin module is deprecated.  "
    "See pyomo.core.base.transformation for Transformation and "
    "TransformationFactory, pyomo.core.base.component for "
    "ModelComponentFactory and pyomo.scripting.interface for "
    "IPyomoScript* interfaces.",
    version='6.0',
)
moved_module(
    "pyomo.core.base.rangeset",
    "pyomo._archive.rangeset",
    msg='The pyomo.core.base.rangeset module is deprecated.  '
    'Import RangeSet objects from pyomo.core.base.set or pyomo.core.',
    version='5.7',
)
moved_module(
    "pyomo.core.base.sets",
    "pyomo._archive.sets",
    msg='The pyomo.core.base.sets module is deprecated.  '
    'Import Set objects from pyomo.core.base.set or pyomo.core.',
    version='5.7',
)
moved_module(
    "pyomo.core.base.template_expr",
    "pyomo._archive.template_expr",
    msg='The pyomo.core.base.template_expr module is deprecated.  '
    'Import expression template objects from pyomo.core.expr.template_expr.',
    version='5.7',
)

relocated_module_attribute(
    'SimpleBlock', 'pyomo.core.base.block.SimpleBlock', version='6.0'
)
relocated_module_attribute('SimpleVar', 'pyomo.core.base.var.SimpleVar', version='6.0')
relocated_module_attribute(
    'SimpleBooleanVar', 'pyomo.core.base.boolean_var.SimpleBooleanVar', version='6.0'
)
# Historically, only a subset of "private" component data classes were imported here
relocated_module_attribute(
    f'_GeneralVarData', f'pyomo.core.base.VarData', version='6.7.2'
)
relocated_module_attribute(
    f'_GeneralBooleanVarData', f'pyomo.core.base.BooleanVarData', version='6.7.2'
)
relocated_module_attribute(
    f'_ExpressionData', f'pyomo.core.base.NamedExpressionData', version='6.7.2'
)
for _cdata in (
    'ConstraintData',
    'LogicalConstraintData',
    'VarData',
    'BooleanVarData',
    'ObjectiveData',
):
    relocated_module_attribute(
        f'_{_cdata}', f'pyomo.core.base.{_cdata}', version='6.7.2'
    )

del _cdata, relocated_module_attribute, moved_module
